#!/usr/bin/env python3
"""
Dynamic Language Loader

Handles loading of Tree-sitter language parsers with efficient caching
and lazy loading for optimal performance.
"""

import importlib
from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:
    from tree_sitter import Language, Parser

try:
    import tree_sitter

    TREE_SITTER_AVAILABLE = True
except ImportError:
    TREE_SITTER_AVAILABLE = False

from .utils import log_warning


class LanguageLoader:
    """Optimized language loader with enhanced caching"""

    # 対応言語とモジュールのマッピング（最適化：frozendict使用を検討）
    LANGUAGE_MODULES = {
        "java": "tree_sitter_java",
        "javascript": "tree_sitter_javascript",
        "typescript": "tree_sitter_typescript",
        "tsx": "tree_sitter_typescript",
        "python": "tree_sitter_python",
        "c": "tree_sitter_c",
        "cpp": "tree_sitter_cpp",
        "rust": "tree_sitter_rust",
        "go": "tree_sitter_go",
        "markdown": "tree_sitter_markdown",
        "sql": "tree_sitter_sql",
        "csharp": "tree_sitter_c_sharp",
        "cs": "tree_sitter_c_sharp",  # C# alias
        # Web languages
        "html": "tree_sitter_html",
        "css": "tree_sitter_css",
        "yaml": "tree_sitter_yaml",
        "yml": "tree_sitter_yaml",  # YAML alias
        # Additional languages
        "php": "tree_sitter_php",
        "ruby": "tree_sitter_ruby",
        "kotlin": "tree_sitter_kotlin",
    }

    # TypeScript特別処理（TypeScriptとTSX）
    TYPESCRIPT_DIALECTS = {"typescript": "typescript", "tsx": "tsx"}

    @property
    def SUPPORTED_LANGUAGES(self) -> list:
        """サポートされている言語のリストを取得するプロパティ"""
        return list(self.LANGUAGE_MODULES.keys())

    def __init__(self) -> None:
        """ローダーを初期化（最適化：事前キャッシュ容量指定）"""
        self._loaded_languages: dict[str, Language] = {}
        self._loaded_modules: dict[str, Any] = {}
        self._availability_cache: dict[str, bool] = {}
        self._parser_cache: dict[str, Parser] = {}  # パーサーキャッシュ追加
        self._unavailable_languages: set[str] = set()  # 利用不可言語の記録

    def is_language_available(self, language: str) -> bool:
        """
        指定された言語のライブラリが利用可能かチェック

        Args:
            language: 言語名

        Returns:
            利用可能性
        """
        # 事前に利用不可とわかっている言語は即座に返す
        if language in self._unavailable_languages:
            return False

        if language in self._availability_cache:
            return self._availability_cache[language]

        if not TREE_SITTER_AVAILABLE:
            self._availability_cache[language] = False
            self._unavailable_languages.add(language)
            return False

        module_name = self.LANGUAGE_MODULES.get(language)
        if not module_name:
            self._availability_cache[language] = False
            self._unavailable_languages.add(language)
            return False

        try:
            importlib.import_module(module_name)
            self._availability_cache[language] = True
            return True
        except ImportError:
            self._availability_cache[language] = False
            self._unavailable_languages.add(language)
            return False

    def load_language(self, language: str) -> Any | None:
        """Load and return a tree-sitter Language object for the specified language"""
        if not TREE_SITTER_AVAILABLE:
            log_warning("Tree-sitter is not available")
            return None

        # キャッシュから取得（最適化）
        if language in self._loaded_languages:
            return self._loaded_languages[language]

        if not self.is_language_available(language):
            return None

        try:
            module_name = self.LANGUAGE_MODULES[language]

            # モジュールキャッシュから取得または新規読み込み
            if module_name not in self._loaded_modules:
                self._loaded_modules[module_name] = importlib.import_module(module_name)

            module = self._loaded_modules[module_name]

            # TypeScript特別処理
            if language in self.TYPESCRIPT_DIALECTS:
                dialect = self.TYPESCRIPT_DIALECTS[language]
                if hasattr(module, "language_typescript") and dialect == "typescript":
                    language_func = module.language_typescript
                elif hasattr(module, "language_tsx") and dialect == "tsx":
                    language_func = module.language_tsx
                elif hasattr(module, "language"):
                    language_func = module.language
                else:
                    return None
            else:
                if hasattr(module, "language"):
                    language_func = module.language
                else:
                    return None

            # Language オブジェクト作成（新しいAPI対応）
            caps_or_lang = language_func()

            # 新しいtree-sitter APIでは、language_func()が直接Languageオブジェクトを返す
            # 古いAPIではPyCapsuleを返すため、適切に処理する
            if hasattr(caps_or_lang, "__class__") and "Language" in str(
                type(caps_or_lang)
            ):
                # 既にLanguageオブジェクトの場合はそのまま使用
                tree_sitter_language = caps_or_lang
            else:
                # PyCapsuleの場合は、Languageオブジェクトを作成
                try:
                    # Use modern tree-sitter API - PyCapsule should be passed to Language constructor
                    tree_sitter_language = tree_sitter.Language(caps_or_lang)
                except Exception as e:
                    log_warning(f"Failed to create Language object for {language}: {e}")
                    return None

            self._loaded_languages[language] = tree_sitter_language
            return tree_sitter_language

        except (ImportError, AttributeError, Exception) as e:
            log_warning(f"Failed to load language '{language}': {e}")
            self._unavailable_languages.add(language)
            return None

    def create_parser_safely(self, language: str) -> Optional["Parser"]:
        """Create a parser for the specified language with error handling"""
        if not TREE_SITTER_AVAILABLE:
            log_warning("Tree-sitter is not available")
            return None

        # パーサーキャッシュから取得
        if language in self._parser_cache:
            return self._parser_cache[language]

        tree_sitter_language = self.load_language(language)
        if tree_sitter_language is None:
            return None

        try:
            # Create parser and set language properly
            parser = tree_sitter.Parser()

            # Ensure we have a proper Language object
            if not hasattr(tree_sitter_language, "__class__") or "Language" not in str(
                type(tree_sitter_language)
            ):
                log_warning(
                    f"Invalid language object for {language}: {type(tree_sitter_language)}"
                )
                return None

            # Set language using the preferred method
            if hasattr(parser, "set_language"):
                parser.set_language(tree_sitter_language)
            elif hasattr(parser, "language"):
                parser.language = tree_sitter_language
            else:
                # Try constructor approach as last resort
                try:
                    parser = tree_sitter.Parser(tree_sitter_language)
                except Exception as e:
                    log_warning(
                        f"Failed to create parser with language constructor for {language}: {e}"
                    )
                    return None

            # Cache and return
            self._parser_cache[language] = parser
            return parser
        except Exception as e:
            log_warning(f"Failed to create parser for '{language}': {e}")
            return None

    def create_parser(self, language: str) -> Optional["Parser"]:
        """Create a parser for the specified language (alias for create_parser_safely)"""
        return self.create_parser_safely(language)

    def get_supported_languages(self) -> list:
        """
        サポートされている言語のリストを取得（最適化：結果キャッシュ）

        Returns:
            サポート言語のリスト
        """
        # 利用可能な言語のみを返す（効率化）
        return [
            lang
            for lang in self.LANGUAGE_MODULES.keys()
            if lang not in self._unavailable_languages
            and self.is_language_available(lang)
        ]

    def clear_cache(self) -> None:
        """キャッシュをクリア（メモリ管理用）"""
        self._loaded_languages.clear()
        self._loaded_modules.clear()
        self._availability_cache.clear()
        self._parser_cache.clear()
        self._unavailable_languages.clear()
        # 可用性キャッシュもクリア
        self._availability_cache.clear()


# グローバルインスタンス（最適化：シングルトンパターン）
_loader_instance = None


def get_loader() -> "LanguageLoader":
    """Get singleton loader instance"""
    global _loader_instance
    if _loader_instance is None:
        _loader_instance = LanguageLoader()
    return _loader_instance


# 後方互換性のため
loader = get_loader()


def check_language_availability(language: str) -> bool:
    """言語の利用可能性をチェック"""
    return get_loader().is_language_available(language)


def create_parser_safely(language: str) -> Optional["Parser"]:
    """安全にパーサーを作成"""
    return get_loader().create_parser_safely(language)


def load_language(language: str) -> Optional["Language"]:
    """言語をロード"""
    return get_loader().load_language(language)



#!/usr/bin/env python3
"""
Path Resolver Utility for MCP Tools

This module provides unified path resolution functionality for all MCP tools,
ensuring consistent handling of relative and absolute paths across different
operating systems.
"""

import logging
import os
from pathlib import Path

logger = logging.getLogger(__name__)


def _normalize_path_cross_platform(path_str: str) -> str:
    """
    Normalize path for cross-platform compatibility.

    Args:
        path_str: Input path string

    Returns:
        Normalized path string
    """
    if not path_str:
        return path_str

    # Handle macOS path normalization issues
    if os.name == "posix":
        # Handle /System/Volumes/Data prefix on macOS
        if path_str.startswith("/System/Volumes/Data/"):
            # Remove the /System/Volumes/Data prefix
            normalized = path_str[len("/System/Volumes/Data") :]
            return normalized

        # Handle /private/var vs /var symlink difference on macOS
        if path_str.startswith("/private/var/"):
            # Always normalize to /var form on macOS for consistency
            var_path = path_str.replace("/private/var/", "/var/", 1)
            return var_path
        elif path_str.startswith("/var/"):
            # Keep /var form as is
            return path_str

    # Handle Windows short path names (8.3 format)
    if os.name == "nt" and path_str:
        try:
            # Try to get the long path name on Windows
            import ctypes
            from ctypes import wintypes

            # GetLongPathNameW function
            _GetLongPathNameW = ctypes.windll.kernel32.GetLongPathNameW
            _GetLongPathNameW.argtypes = [
                wintypes.LPCWSTR,
                wintypes.LPWSTR,
                wintypes.DWORD,
            ]
            _GetLongPathNameW.restype = wintypes.DWORD

            # Buffer for the long path
            buffer_size = 1000
            buffer = ctypes.create_unicode_buffer(buffer_size)

            # Get the long path name
            result = _GetLongPathNameW(path_str, buffer, buffer_size)
            if result > 0 and result < buffer_size:
                long_path = buffer.value
                if long_path and long_path != path_str:
                    return long_path
        except (ImportError, AttributeError, OSError):
            # If Windows API calls fail, continue with original path
            pass

    return path_str


def _is_windows_absolute_path(path_str: str) -> bool:
    """
    Check if a path is a Windows-style absolute path.

    Args:
        path_str: Path string to check

    Returns:
        True if it's a Windows absolute path (e.g., C:\\path or C:/path)
    """
    if not path_str or len(path_str) < 3:
        return False

    # Check for drive letter pattern: X:\ or X:/
    return path_str[1] == ":" and path_str[0].isalpha() and path_str[2] in ("\\", "/")


class PathResolver:
    """
    Utility class for resolving file paths in MCP tools.

    Handles relative path resolution against project root and provides
    cross-platform compatibility for Windows, macOS, and Linux.
    """

    def __init__(self, project_root: str | None = None):
        """
        Initialize the path resolver.

        Args:
            project_root: Optional project root directory for resolving relative paths
        """
        self.project_root = None
        self._cache: dict[str, str] = {}  # Simple cache for resolved paths
        self._cache_size_limit = 100  # Limit cache size to prevent memory issues

        if project_root:
            # Use pathlib for consistent path handling, but preserve relative paths for compatibility
            path_obj = Path(project_root)
            if path_obj.is_absolute():
                resolved_root = str(path_obj.resolve())
                # Apply cross-platform normalization
                self.project_root = _normalize_path_cross_platform(resolved_root)
            else:
                # For relative paths, normalize but don't resolve to absolute
                self.project_root = str(path_obj)
            logger.debug(
                f"PathResolver initialized with project root: {self.project_root}"
            )

    def resolve(self, file_path: str) -> str:
        """
        Resolve a file path to an absolute path.

        Args:
            file_path: Input file path (can be relative or absolute)

        Returns:
            Resolved absolute file path

        Raises:
            ValueError: If file_path is empty or None
            TypeError: If file_path is not a string
        """
        if not file_path:
            raise ValueError("file_path cannot be empty or None")

        if not isinstance(file_path, str):
            raise TypeError(
                f"file_path must be a string, got {type(file_path).__name__}"
            )

        # Check cache first
        if file_path in self._cache:
            logger.debug(f"Cache hit for path: {file_path}")
            return self._cache[file_path]

        # Normalize path separators first
        normalized_input = file_path.replace("\\", "/")

        # Special handling for Windows absolute paths on non-Windows systems
        if _is_windows_absolute_path(file_path) and os.name != "nt":
            # On non-Windows systems, Windows absolute paths should be treated as-is
            # Don't try to resolve them relative to project root
            logger.debug(f"Windows absolute path on non-Windows system: {file_path}")
            self._add_to_cache(file_path, file_path)
            return file_path

        path_obj = Path(normalized_input)

        # Handle Unix-style absolute paths on Windows (starting with /) FIRST
        # This must come before the is_absolute() check because Unix paths aren't
        # considered absolute on Windows by pathlib
        if (
            normalized_input.startswith("/") and os.name == "nt"
        ):  # Check if we're on Windows
            # On Windows, convert Unix-style absolute paths to Windows format
            # by prepending the current drive with proper separator
            current_drive = Path.cwd().drive
            if current_drive:
                # Remove leading slash and join with current drive
                unix_path_without_slash = normalized_input[1:]
                # Ensure proper Windows path format with backslash after drive
                resolved_path = str(
                    Path(current_drive + "\\") / unix_path_without_slash
                )
                logger.debug(
                    f"Converted Unix absolute path: {file_path} -> {resolved_path}"
                )
                # Apply cross-platform normalization
                resolved_path = _normalize_path_cross_platform(resolved_path)
                self._add_to_cache(file_path, resolved_path)
                return resolved_path
            # If no drive available, continue with normal processing

        # Check if path is absolute
        if path_obj.is_absolute():
            resolved_path = str(path_obj.resolve())
            logger.debug(f"Path already absolute: {file_path} -> {resolved_path}")
            # Apply cross-platform normalization
            resolved_path = _normalize_path_cross_platform(resolved_path)
            self._add_to_cache(file_path, resolved_path)
            return resolved_path

        # If we have a project root, resolve relative to it
        if self.project_root:
            resolved_path = str((Path(self.project_root) / normalized_input).resolve())
            logger.debug(
                f"Resolved relative path '{file_path}' to '{resolved_path}' using project root"
            )
            # Apply cross-platform normalization
            resolved_path = _normalize_path_cross_platform(resolved_path)
            self._add_to_cache(file_path, resolved_path)
            return resolved_path

        # Fallback: try to resolve relative to current working directory
        resolved_path = str(Path(normalized_input).resolve())
        logger.debug(
            f"Resolved relative path '{file_path}' to '{resolved_path}' using current working directory"
        )

        # Apply cross-platform normalization
        resolved_path = _normalize_path_cross_platform(resolved_path)

        # Cache the result
        self._add_to_cache(file_path, resolved_path)

        return resolved_path

    def _add_to_cache(self, file_path: str, resolved_path: str) -> None:
        """
        Add a resolved path to the cache.

        Args:
            file_path: Original file path
            resolved_path: Resolved absolute path
        """
        # Limit cache size to prevent memory issues
        if len(self._cache) >= self._cache_size_limit:
            # Remove oldest entries (simple FIFO)
            oldest_key = next(iter(self._cache))
            del self._cache[oldest_key]
            logger.debug(f"Cache full, removed oldest entry: {oldest_key}")

        self._cache[file_path] = resolved_path
        logger.debug(f"Cached path resolution: {file_path} -> {resolved_path}")

    def clear_cache(self) -> None:
        """Clear the path resolution cache."""
        cache_size = len(self._cache)
        self._cache.clear()
        logger.info(f"Cleared path resolution cache ({cache_size} entries)")

    def get_cache_stats(self) -> dict[str, int]:
        """
        Get cache statistics.

        Returns:
            Dictionary with cache statistics
        """
        return {"size": len(self._cache), "limit": self._cache_size_limit}

    def is_relative(self, file_path: str) -> bool:
        """
        Check if a file path is relative.

        Args:
            file_path: File path to check

        Returns:
            True if the path is relative, False if absolute
        """
        return not Path(file_path).is_absolute()

    def get_relative_path(self, absolute_path: str) -> str:
        """
        Get the relative path from project root to the given absolute path.

        Args:
            absolute_path: Absolute file path

        Returns:
            Relative path from project root, or the original path if no project root

        Raises:
            ValueError: If absolute_path is not actually absolute
        """
        abs_path = Path(absolute_path)
        if not abs_path.is_absolute():
            raise ValueError(f"Path is not absolute: {absolute_path}")

        if not self.project_root:
            return absolute_path

        try:
            # Get relative path from project root using pathlib
            project_path = Path(self.project_root)

            # Normalize both paths for consistent comparison
            normalized_abs_path = _normalize_path_cross_platform(
                str(abs_path.resolve())
            )
            normalized_project_root = _normalize_path_cross_platform(
                str(project_path.resolve())
            )

            # Convert back to Path objects for relative_to calculation
            normalized_abs_path_obj = Path(normalized_abs_path)
            normalized_project_root_obj = Path(normalized_project_root)

            relative_path = str(
                normalized_abs_path_obj.relative_to(normalized_project_root_obj)
            )
            logger.debug(
                f"Converted absolute path '{absolute_path}' to relative path '{relative_path}'"
            )
            return relative_path
        except ValueError:
            # Paths are on different drives (Windows) or other error
            logger.warning(
                f"Could not convert absolute path '{absolute_path}' to relative path"
            )
            return absolute_path

    def validate_path(self, file_path: str) -> tuple[bool, str | None]:
        """
        Validate if a file path is valid and safe.

        Args:
            file_path: File path to validate

        Returns:
            Tuple of (is_valid, error_message)
        """
        try:
            resolved_path = self.resolve(file_path)
            resolved_path_obj = Path(resolved_path)

            if not resolved_path_obj.exists():
                return False, f"File does not exist: {resolved_path}"

            # Check if it's a file (not directory)
            if not resolved_path_obj.is_file():
                return False, f"Path is not a file: {resolved_path}"

            # Check if it's a symlink (reject symlinks for security)
            try:
                if resolved_path_obj.is_symlink():
                    return False, f"Path is a symlink: {resolved_path}"
            except (OSError, AttributeError):
                # is_symlink() might not be available on all platforms
                # or might fail due to permissions, skip this check
                pass

            # Check if it's within project root (if we have one)
            if self.project_root:
                try:
                    project_path = Path(self.project_root).resolve()
                    resolved_abs_path = resolved_path_obj.resolve()
                    # Check if the resolved path is within the project root
                    resolved_abs_path.relative_to(project_path)
                except ValueError:
                    return False, f"File path is outside project root: {resolved_path}"

            return True, None

        except Exception as e:
            return False, f"Path validation error: {str(e)}"

    def get_project_root(self) -> str | None:
        """
        Get the current project root.

        Returns:
            Project root path or None if not set
        """
        return self.project_root

    def set_project_root(self, project_root: str) -> None:
        """
        Set or update the project root.

        Args:
            project_root: New project root directory
        """
        if project_root:
            # Use pathlib for consistent path handling, but preserve relative paths for compatibility
            path_obj = Path(project_root)
            if path_obj.is_absolute():
                resolved_root = str(path_obj.resolve())
                # Apply cross-platform normalization
                self.project_root = _normalize_path_cross_platform(resolved_root)
            else:
                # For relative paths, normalize but don't resolve to absolute
                self.project_root = str(path_obj)
            logger.info(f"Project root updated to: {self.project_root}")
        else:
            self.project_root = None
            logger.info("Project root cleared")


# Convenience function for backward compatibility
def resolve_path(file_path: str, project_root: str | None = None) -> str:
    """
    Convenience function to resolve a file path.

    Args:
        file_path: File path to resolve
        project_root: Optional project root directory

    Returns:
        Resolved absolute file path
    """
    resolver = PathResolver(project_root)
    return resolver.resolve(file_path)



#!/usr/bin/env python3
"""
Project Statistics Resource for MCP

This module provides MCP resource implementation for accessing project
statistics and analysis data. The resource allows dynamic access to
project analysis results through URI-based identification.
"""

import json
import logging
import re
from datetime import datetime
from pathlib import Path
from typing import Any, cast

from tree_sitter_analyzer.core.analysis_engine import (
    AnalysisRequest,
    get_analysis_engine,
)
from tree_sitter_analyzer.language_detector import (
    detect_language_from_file,
    is_language_supported,
)

logger = logging.getLogger(__name__)


class ProjectStatsResource:
    """
    MCP resource for accessing project statistics and analysis data

    This resource provides access to project analysis results through the MCP protocol.
    It supports various types of statistics including overview, language breakdown,
    complexity metrics, and file-level information.

    URI Format: code://stats/{stats_type}

    Supported stats types:
        - overview: General project overview
        - languages: Language breakdown and statistics
        - complexity: Complexity metrics and analysis
        - files: File-level statistics and information

    Examples:
        - code://stats/overview
        - code://stats/languages
        - code://stats/complexity
        - code://stats/files
    """

    def __init__(self) -> None:
        """Initialize the project statistics resource"""
        self._uri_pattern = re.compile(r"^code://stats/(.+)$")
        self._project_path: str | None = None
        self.analysis_engine = get_analysis_engine()
        # Use unified analysis engine instead of deprecated AdvancedAnalyzer

        # Supported statistics types
        self._supported_stats_types = {"overview", "languages", "complexity", "files"}

    @property
    def project_root(self) -> str | None:
        """Get the current project root path"""
        return self._project_path

    @project_root.setter
    def project_root(self, value: str | None) -> None:
        """Set the current project root path"""
        self._project_path = value

    def get_resource_info(self) -> dict[str, Any]:
        """
        Get resource information for MCP registration

        Returns:
            Dict containing resource metadata
        """
        return {
            "name": "project_stats",
            "description": "Access to project statistics and analysis data",
            "uri_template": "code://stats/{stats_type}",
            "mime_type": "application/json",
        }

    def matches_uri(self, uri: str) -> bool:
        """
        Check if the URI matches this resource pattern

        Args:
            uri: The URI to check

        Returns:
            True if the URI matches the project stats pattern
        """
        return bool(self._uri_pattern.match(uri))

    def _extract_stats_type(self, uri: str) -> str:
        """
        Extract statistics type from URI

        Args:
            uri: The URI to extract stats type from

        Returns:
            The extracted statistics type

        Raises:
            ValueError: If URI format is invalid
        """
        match = self._uri_pattern.match(uri)
        if not match:
            raise ValueError(f"Invalid URI format: {uri}")

        return match.group(1)

    def set_project_path(self, project_path: str) -> None:
        """
        Set the project path for analysis

        Args:
            project_path: Path to the project directory

        Raises:
            TypeError: If project_path is not a string
            ValueError: If project_path is empty
        """
        if not isinstance(project_path, str):
            raise TypeError("Project path must be a string")

        if not project_path:
            raise ValueError("Project path cannot be empty")

        self._project_path = project_path

        # Note: analysis_engine is already initialized in __init__
        # No need to reinitialize here

        logger.debug(f"Set project path to: {project_path}")

    def _validate_project_path(self) -> None:
        """
        Validate that project path is set and exists

        Raises:
            ValueError: If project path is not set
            FileNotFoundError: If project path doesn't exist
        """
        if not self._project_path:
            raise ValueError("Project path not set. Call set_project_path() first.")

        if self._project_path is None:
            raise ValueError("Project path is not set")
        project_dir = Path(self._project_path)
        if not project_dir.exists():
            raise FileNotFoundError(
                f"Project directory does not exist: {self._project_path}"
            )

        if not project_dir.is_dir():
            raise FileNotFoundError(
                f"Project path is not a directory: {self._project_path}"
            )

    def _is_supported_code_file(self, file_path: Path) -> bool:
        """
        Check if the file is a supported code file using language detection

        Args:
            file_path: Path to the file

        Returns:
            True if the file is a supported code file
        """
        try:
            language = detect_language_from_file(str(file_path))
            return is_language_supported(language)
        except Exception:
            return False

    def _get_language_from_file(self, file_path: Path) -> str:
        """
        Get language from file using language detector

        Args:
            file_path: Path to the file

        Returns:
            Detected language name
        """
        try:
            return detect_language_from_file(str(file_path))
        except Exception:
            return "unknown"

    async def _generate_overview_stats(self) -> dict[str, Any]:
        """
        Generate overview statistics for the project

        Returns:
            Dictionary containing overview statistics
        """
        logger.debug("Generating overview statistics")

        # Scan project directory for actual file counts
        if self._project_path is None:
            raise ValueError("Project path is not set")
        project_dir = Path(self._project_path)
        total_files = 0
        total_lines = 0
        language_counts: dict[str, int] = {}

        for file_path in project_dir.rglob("*"):
            if file_path.is_file() and self._is_supported_code_file(file_path):
                total_files += 1
                try:
                    from ...encoding_utils import read_file_safe

                    content, _ = read_file_safe(file_path)
                    file_lines = len(content.splitlines())
                    total_lines += file_lines
                except Exception as e:
                    logger.debug(
                        f"Skipping unreadable file during overview scan: {file_path} ({e})"
                    )
                    continue
                language = self._get_language_from_file(file_path)
                if language != "unknown":
                    language_counts[language] = language_counts.get(language, 0) + 1

        analysis_result = {
            "total_files": total_files,
            "total_lines": total_lines,
            "languages": [
                {"name": lang, "file_count": count}
                for lang, count in language_counts.items()
            ],
        }

        # Extract overview information
        languages_data = analysis_result.get("languages", [])
        if languages_data is None:
            languages_data = []

        # Ensure languages_data is a list for iteration
        if not isinstance(languages_data, list):
            languages_data = []

        overview = {
            "total_files": analysis_result.get("total_files", 0),
            "total_lines": analysis_result.get("total_lines", 0),
            "languages": [
                str(lang["name"])
                for lang in languages_data
                if isinstance(lang, dict) and "name" in lang
            ],
            "project_path": self._project_path,
            "last_updated": datetime.now().isoformat(),
        }

        logger.debug(f"Generated overview with {overview['total_files']} files")
        return overview

    async def _generate_languages_stats(self) -> dict[str, Any]:
        """
        Generate language-specific statistics

        Returns:
            Dictionary containing language statistics
        """
        logger.debug("Generating language statistics")

        # Scan project directory for actual language counts
        if self._project_path is None:
            raise ValueError("Project path is not set")
        project_dir = Path(self._project_path)
        total_files = 0
        total_lines = 0
        language_data = {}

        for file_path in project_dir.rglob("*"):
            if file_path.is_file() and self._is_supported_code_file(file_path):
                total_files += 1
                try:
                    from ...encoding_utils import read_file_safe

                    content, _ = read_file_safe(file_path)
                    file_lines = len(content.splitlines())
                    total_lines += file_lines
                except Exception as e:
                    logger.debug(f"Failed to count lines for {file_path}: {e}")
                    file_lines = 0

                language = self._get_language_from_file(file_path)
                if language != "unknown":
                    if language not in language_data:
                        language_data[language] = {"file_count": 0, "line_count": 0}
                    language_data[language]["file_count"] += 1
                    language_data[language]["line_count"] += file_lines

        # Convert to list format and calculate percentages
        languages_list = []
        for lang, data in language_data.items():
            percentage = (
                round((data["line_count"] / total_lines) * 100, 2)
                if total_lines > 0
                else 0.0
            )
            languages_list.append(
                {
                    "name": lang,
                    "file_count": data["file_count"],
                    "line_count": data["line_count"],
                    "percentage": percentage,
                }
            )

        languages_stats = {
            "languages": languages_list,
            "total_languages": len(languages_list),
            "last_updated": datetime.now().isoformat(),
        }

        logger.debug(f"Generated stats for {len(languages_list)} languages")
        return languages_stats

    async def _generate_complexity_stats(self) -> dict[str, Any]:
        """
        Generate complexity statistics

        Returns:
            Dictionary containing complexity statistics
        """
        logger.debug("Generating complexity statistics")

        # Analyze files for complexity
        if self._project_path is None:
            raise ValueError("Project path is not set")
        project_dir = Path(self._project_path)
        complexity_data = []
        total_complexity = 0
        max_complexity = 0
        file_count = 0

        # Analyze each supported code file
        for file_path in project_dir.rglob("*"):
            if file_path.is_file() and self._is_supported_code_file(file_path):
                try:
                    language = self._get_language_from_file(file_path)

                    # Use appropriate analyzer based on language
                    if language == "java":
                        # Use analysis engine for Java
                        file_analysis = await self.analysis_engine.analyze_file(
                            str(file_path)
                        )
                        if file_analysis and hasattr(file_analysis, "methods"):
                            # Extract complexity from methods if available
                            complexity = sum(
                                method.complexity_score or 0
                                for method in file_analysis.methods
                            )
                        elif file_analysis and hasattr(file_analysis, "elements"):
                            # Extract complexity from elements for new architecture
                            methods = [
                                e
                                for e in file_analysis.elements
                                if hasattr(e, "complexity_score")
                            ]
                            complexity = sum(
                                getattr(method, "complexity_score", 0) or 0
                                for method in methods
                            )
                        else:
                            complexity = 0
                    else:
                        # Use universal analyzer for other languages
                        request = AnalysisRequest(
                            file_path=str(file_path), language=language
                        )
                        file_analysis_result = await self.analysis_engine.analyze(
                            request
                        )

                        complexity = 0
                        if file_analysis_result and file_analysis_result.success:
                            analysis_dict = file_analysis_result.to_dict()
                            # Assuming complexity is part of the metrics in the new structure
                            if (
                                "metrics" in analysis_dict
                                and "complexity" in analysis_dict["metrics"]
                            ):
                                complexity = analysis_dict["metrics"]["complexity"].get(
                                    "total", 0
                                )

                    if complexity > 0:
                        complexity_data.append(
                            {
                                "file": str(file_path.relative_to(project_dir)),
                                "language": language,
                                "complexity": complexity,
                            }
                        )

                        total_complexity += complexity
                        max_complexity = max(max_complexity, complexity)
                        file_count += 1

                except Exception as e:
                    logger.warning(f"Failed to analyze complexity for {file_path}: {e}")
                    continue

        # Calculate average complexity
        avg_complexity = total_complexity / file_count if file_count > 0 else 0

        complexity_stats = {
            "average_complexity": round(avg_complexity, 2),
            "max_complexity": max_complexity,
            "total_files_analyzed": file_count,
            "files_by_complexity": sorted(
                complexity_data,
                key=lambda x: int(cast(int, x.get("complexity", 0))),
                reverse=True,
            ),
            "last_updated": datetime.now().isoformat(),
        }

        logger.debug(f"Generated complexity stats for {file_count} files")
        return complexity_stats

    async def _generate_files_stats(self) -> dict[str, Any]:
        """
        Generate file-level statistics

        Returns:
            Dictionary containing file statistics
        """
        logger.debug("Generating file statistics")

        # Get detailed file information
        files_data = []
        if self._project_path is None:
            raise ValueError("Project path is not set")
        project_dir = Path(self._project_path)

        # Analyze each supported code file
        for file_path in project_dir.rglob("*"):
            if file_path.is_file() and self._is_supported_code_file(file_path):
                try:
                    # Get file stats
                    file_stats = file_path.stat()

                    # Determine language using language detector
                    language = self._get_language_from_file(file_path)

                    # Count lines
                    try:
                        from ...encoding_utils import read_file_safe

                        content, _ = read_file_safe(file_path)
                        line_count = len(content.splitlines())
                    except Exception:
                        line_count = 0

                    files_data.append(
                        {
                            "path": str(file_path.relative_to(project_dir)),
                            "language": language,
                            "line_count": line_count,
                            "size_bytes": file_stats.st_size,
                            "modified": datetime.fromtimestamp(
                                file_stats.st_mtime
                            ).isoformat(),
                        }
                    )

                except Exception as e:
                    logger.warning(f"Failed to get stats for {file_path}: {e}")
                    continue

        files_stats = {
            "files": sorted(
                files_data,
                key=lambda x: int(cast(int, x.get("line_count", 0))),
                reverse=True,
            ),
            "total_count": len(files_data),
            "last_updated": datetime.now().isoformat(),
        }

        logger.debug(f"Generated stats for {len(files_data)} files")
        return files_stats

    async def read_resource(self, uri: str) -> str:
        """
        Read resource content from URI

        Args:
            uri: The resource URI to read

        Returns:
            Resource content as JSON string

        Raises:
            ValueError: If URI format is invalid or stats type is unsupported
            FileNotFoundError: If project path doesn't exist
        """
        logger.debug(f"Reading resource: {uri}")

        # Validate URI format
        if not self.matches_uri(uri):
            raise ValueError(f"URI does not match project stats pattern: {uri}")

        # Extract statistics type
        stats_type = self._extract_stats_type(uri)

        # Validate statistics type
        if stats_type not in self._supported_stats_types:
            raise ValueError(
                f"Unsupported statistics type: {stats_type}. "
                f"Supported types: {', '.join(self._supported_stats_types)}"
            )

        # Validate project path
        self._validate_project_path()

        # Generate statistics based on type
        try:
            if stats_type == "overview":
                stats_data = await self._generate_overview_stats()
            elif stats_type == "languages":
                stats_data = await self._generate_languages_stats()
            elif stats_type == "complexity":
                stats_data = await self._generate_complexity_stats()
            elif stats_type == "files":
                stats_data = await self._generate_files_stats()
            else:
                raise ValueError(f"Unknown statistics type: {stats_type}")

            # Convert to JSON
            json_content = json.dumps(stats_data, indent=2, ensure_ascii=False)
            logger.debug(f"Successfully generated {stats_type} statistics")
            return json_content

        except Exception as e:
            logger.error(f"Failed to generate {stats_type} statistics: {e}")
            raise

    def get_supported_schemes(self) -> list[str]:
        """
        Get list of supported URI schemes

        Returns:
            List of supported schemes
        """
        return ["code"]

    def get_supported_resource_types(self) -> list[str]:
        """
        Get list of supported resource types

        Returns:
            List of supported resource types
        """
        return ["stats"]

    def get_supported_stats_types(self) -> list[str]:
        """
        Get list of supported statistics types

        Returns:
            List of supported statistics types
        """
        return list(self._supported_stats_types)

    def __str__(self) -> str:
        """String representation of the resource"""
        return "ProjectStatsResource(pattern=code://stats/{stats_type})"

    def __repr__(self) -> str:
        """Detailed string representation of the resource"""
        return (
            f"ProjectStatsResource(uri_pattern={self._uri_pattern.pattern}, "
            f"project_path={self._project_path})"
        )



#!/usr/bin/env python3
"""
JavaScript-specific table formatter.

Provides specialized formatting for JavaScript code analysis results,
handling modern JavaScript features like ES6+ syntax, async/await,
classes, modules, and framework-specific patterns.
"""

from typing import Any

from .base_formatter import BaseTableFormatter


class JavaScriptTableFormatter(BaseTableFormatter):
    """Table formatter specialized for JavaScript"""

    def format(self, data: dict[str, Any] | None, format_type: str = "full") -> str:
        """Format data using the configured format type"""
        # Handle None data gracefully
        if data is None:
            data = {}

        # Ensure data is a dictionary
        if not isinstance(data, dict):
            raise TypeError(f"Expected dict, got {type(data)}")

        if format_type:
            # Check for supported format types
            supported_formats = ["full", "compact", "csv", "json"]
            if format_type not in supported_formats:
                raise ValueError(
                    f"Unsupported format type: {format_type}. Supported formats: {supported_formats}"
                )

            # Handle json format separately
            if format_type == "json":
                return self._format_json(data)

            # Temporarily change format type for this call
            original_format = self.format_type
            self.format_type = format_type
            result = self.format_structure(data)
            self.format_type = original_format
            return result
        return self.format_structure(data)

    def _format_full_table(self, data: dict[str, Any]) -> str:
        """Full table format for JavaScript"""
        if not isinstance(data, dict):
            raise TypeError(f"Expected dict, got {type(data)}")

        lines = []

        # Header - JavaScript (module/file based)
        file_path = data.get("file_path", "Unknown")
        if file_path is None:
            file_path = "Unknown"
        file_name = str(file_path).split("/")[-1].split("\\")[-1]
        module_name = (
            file_name.replace(".js", "").replace(".jsx", "").replace(".mjs", "")
        )

        # Check if this is a module (has exports)
        exports = data.get("exports", [])
        if exports is None:
            exports = []
        is_module = len(exports) > 0

        if is_module:
            lines.append(f"# Module: {module_name}")
        else:
            lines.append(f"# Script: {module_name}")
        lines.append("")

        # Imports
        imports = data.get("imports", [])
        if imports:
            lines.append("## Imports")
            lines.append("```javascript")
            for imp in imports:
                if isinstance(imp, str):
                    # Handle malformed data where import is a string
                    import_statement = imp
                elif isinstance(imp, dict):
                    import_statement = imp.get("statement", "")
                    if not import_statement:
                        # Construct import statement from parts
                        source = imp.get("source", "")
                        name = imp.get("name", "")
                        if name and source:
                            import_statement = f"import {name} from {source};"
                else:
                    import_statement = str(imp)
                lines.append(import_statement)
            lines.append("```")
            lines.append("")

        # Module Info
        stats = data.get("statistics", {})
        if stats is None or not isinstance(stats, dict):
            stats = {}
        classes = data.get("classes", [])
        if classes is None:
            classes = []

        lines.append("## Module Info")
        lines.append("| Property | Value |")
        lines.append("|----------|-------|")
        lines.append(f"| File | {file_name} |")
        lines.append(f"| Type | {'ES6 Module' if is_module else 'Script'} |")
        lines.append(f"| Functions | {stats.get('function_count', 0)} |")
        lines.append(f"| Classes | {len(classes)} |")
        lines.append(f"| Variables | {stats.get('variable_count', 0)} |")
        lines.append(f"| Exports | {len(exports)} |")
        lines.append("")

        # Classes (if any)
        if classes:
            lines.append("## Classes")
            lines.append("| Class | Type | Extends | Lines | Methods | Properties |")
            lines.append("|-------|------|---------|-------|---------|------------|")

            for class_info in classes:
                name = str(class_info.get("name", "Unknown"))
                class_type = "class"  # JavaScript only has classes
                extends = str(class_info.get("superclass", "")) or "-"
                line_range = class_info.get("line_range", {})
                lines_str = f"{line_range.get('start', 0)}-{line_range.get('end', 0)}"

                # Count methods within the class
                class_methods = [
                    m
                    for m in data.get("methods", [])
                    if line_range.get("start", 0)
                    <= m.get("line_range", {}).get("start", 0)
                    <= line_range.get("end", 0)
                ]

                # Count properties (class fields)
                class_properties = [
                    v
                    for v in data.get("variables", [])
                    if line_range.get("start", 0)
                    <= v.get("line_range", {}).get("start", 0)
                    <= line_range.get("end", 0)
                ]

                lines.append(
                    f"| {name} | {class_type} | {extends} | {lines_str} | {len(class_methods)} | {len(class_properties)} |"
                )
            lines.append("")

        # Variables/Constants
        variables = data.get("variables", [])
        if variables:
            lines.append("## Variables")
            lines.append("| Name | Type | Scope | Kind | Line | Value |")
            lines.append("|------|------|-------|------|------|-------|")

            for var in variables:
                name = str(var.get("name", ""))
                # Try to get value from initializer or value field
                var_value = var.get("initializer") or var.get("value", "")
                var_type = self._infer_js_type(var_value)
                scope = self._determine_scope(var)
                kind = self._get_variable_kind(var)
                line = var.get("line_range", {}).get("start", 0)
                value = str(var_value)[:30] + (
                    "..." if len(str(var_value)) > 30 else ""
                )
                value = value.replace("\n", " ").replace("|", "\\|")

                lines.append(
                    f"| {name} | {var_type} | {scope} | {kind} | {line} | {value} |"
                )
            lines.append("")

        # Functions
        functions = data.get("functions", [])
        if functions:
            # Group functions by type
            regular_functions = [
                f
                for f in functions
                if not self._is_method(f) and not f.get("is_async", False)
            ]
            async_functions = [
                f
                for f in functions
                if not self._is_method(f) and f.get("is_async", False)
            ]
            methods = [f for f in functions if self._is_method(f)]

            # Regular Functions
            if regular_functions:
                lines.append("## Functions")
                lines.append(
                    "| Function | Parameters | Type | Lines | Complexity | JSDoc |"
                )
                lines.append(
                    "|----------|------------|------|-------|------------|-------|"
                )

                for func in regular_functions:
                    lines.append(self._format_function_row(func))
                lines.append("")

            # Async Functions
            if async_functions:
                lines.append("## Async Functions")
                lines.append(
                    "| Function | Parameters | Type | Lines | Complexity | JSDoc |"
                )
                lines.append(
                    "|----------|------------|------|-------|------------|-------|"
                )

                for func in async_functions:
                    lines.append(self._format_function_row(func))
                lines.append("")

            # Methods (class methods)
            if methods:
                lines.append("## Methods")
                lines.append(
                    "| Method | Class | Parameters | Type | Lines | Complexity | JSDoc |"
                )
                lines.append(
                    "|--------|-------|------------|------|-------|------------|-------|"
                )

                for method in methods:
                    lines.append(self._format_method_row(method))
                lines.append("")

        # Exports
        if exports:
            lines.append("## Exports")
            lines.append("| Export | Type | Name | Default |")
            lines.append("|--------|------|------|---------|")

            # Handle malformed exports data
            if isinstance(exports, list):
                for export in exports:
                    try:
                        export_type = self._get_export_type(export)
                        if isinstance(export, dict):
                            name = str(export.get("name", ""))
                            is_default = "✓" if export.get("is_default", False) else "-"
                        else:
                            name = str(export)
                            is_default = "-"
                            export_type = "unknown"

                        lines.append(f"| {export_type} | {name} | {is_default} |")
                    except (TypeError, AttributeError):
                        # Handle malformed export data gracefully
                        lines.append(f"| unknown | {str(export)} | - |")
            else:
                # Handle case where exports is not a list (malformed data)
                lines.append(f"| unknown | {str(exports)} | - |")
            lines.append("")

        # Trim trailing blank lines
        while lines and lines[-1] == "":
            lines.pop()

        return "\n".join(lines)

    def _format_compact_table(self, data: dict[str, Any]) -> str:
        """Compact table format for JavaScript"""
        lines = []

        # Header
        file_path = data.get("file_path", "Unknown")
        file_name = file_path.split("/")[-1].split("\\")[-1]
        module_name = (
            file_name.replace(".js", "").replace(".jsx", "").replace(".mjs", "")
        )
        lines.append(f"# {module_name}")
        lines.append("")

        # Info
        stats = data.get("statistics", {})
        lines.append("## Info")
        lines.append("| Property | Value |")
        lines.append("|----------|-------|")
        lines.append(f"| Functions | {stats.get('function_count', 0)} |")
        lines.append(f"| Classes | {len(data.get('classes', []))} |")
        lines.append(f"| Variables | {stats.get('variable_count', 0)} |")
        lines.append(f"| Exports | {len(data.get('exports', []))} |")
        lines.append("")

        # Functions (compact)
        functions = data.get("functions", [])
        if functions:
            lines.append("## Functions")
            lines.append("| Function | Params | Type | L | Cx | Doc |")
            lines.append("|----------|--------|------|---|----|----|")

            for func in functions:
                name = str(func.get("name", ""))
                params = self._create_compact_params(func)
                func_type = self._get_function_type_short(func)
                line_range = func.get("line_range", {})
                lines_str = f"{line_range.get('start', 0)}-{line_range.get('end', 0)}"
                complexity = func.get("complexity_score", 0)
                doc = self._clean_csv_text(
                    self._extract_doc_summary(str(func.get("jsdoc", "")))
                )

                lines.append(
                    f"| {name} | {params} | {func_type} | {lines_str} | {complexity} | {doc} |"
                )
            lines.append("")

        # Trim trailing blank lines
        while lines and lines[-1] == "":
            lines.pop()

        return "\n".join(lines)

    def _format_function_row(self, func: dict[str, Any]) -> str:
        """Format a function table row for JavaScript"""
        name = str(func.get("name", ""))
        params = self._create_full_params(func)
        func_type = self._get_function_type(func)
        line_range = func.get("line_range", {})
        lines_str = f"{line_range.get('start', 0)}-{line_range.get('end', 0)}"
        complexity = func.get("complexity_score", 0)
        doc = self._clean_csv_text(
            self._extract_doc_summary(str(func.get("jsdoc", "")))
        )

        return (
            f"| {name} | {params} | {func_type} | {lines_str} | {complexity} | {doc} |"
        )

    def _format_method_row(self, method: dict[str, Any]) -> str:
        """Format a method table row for JavaScript"""
        name = str(method.get("name", ""))
        class_name = self._get_method_class(method)
        params = self._create_full_params(method)
        method_type = self._get_method_type(method)
        line_range = method.get("line_range", {})
        lines_str = f"{line_range.get('start', 0)}-{line_range.get('end', 0)}"
        complexity = method.get("complexity_score", 0)
        doc = self._clean_csv_text(
            self._extract_doc_summary(str(method.get("jsdoc", "")))
        )

        return f"| {name} | {class_name} | {params} | {method_type} | {lines_str} | {complexity} | {doc} |"

    def _create_full_params(self, func: dict[str, Any]) -> str:
        """Create full parameter list for JavaScript functions"""
        params = func.get("parameters", [])
        if not params:
            return "()"

        # Handle malformed data where parameters might be a string
        if isinstance(params, str):
            # If parameters is a malformed string, return empty params
            return "()"

        param_strs = []
        for param in params:
            if isinstance(param, dict):
                param_name = param.get("name", "")
                param_type = param.get("type", "")
                if param_type:
                    param_strs.append(f"{param_name}: {param_type}")
                else:
                    param_strs.append(param_name)
            else:
                param_strs.append(str(param))

        params_str = ", ".join(param_strs)
        if len(params_str) > 50:
            params_str = params_str[:47] + "..."

        return f"({params_str})"

    def _create_compact_params(self, func: dict[str, Any]) -> str:
        """Create compact parameter list for JavaScript functions"""
        params = func.get("parameters", [])
        if not params:
            return "()"

        # Handle malformed data where parameters might be a string
        if isinstance(params, str):
            # If parameters is a malformed string, return empty params
            return "()"

        param_count = len(params)
        if param_count <= 3:
            param_names = [
                param.get("name", str(param)) if isinstance(param, dict) else str(param)
                for param in params
            ]
            return f"({','.join(param_names)})"
        else:
            return f"({param_count} params)"

    def _get_function_type(self, func: dict[str, Any]) -> str:
        """Get full function type for JavaScript"""
        if func.get("is_async", False):
            return "async function"
        elif func.get("is_generator", False):
            return "generator"
        elif func.get("is_arrow", False):
            return "arrow"
        elif self._is_method(func):
            if func.get("is_constructor", False):
                return "constructor"
            elif func.get("is_getter", False):
                return "getter"
            elif func.get("is_setter", False):
                return "setter"
            elif func.get("is_static", False):
                return "static method"
            else:
                return "method"
        else:
            return "function"

    def _get_function_type_short(self, func: dict[str, Any]) -> str:
        """Get short function type for JavaScript"""
        if func.get("is_async", False):
            return "async"
        elif func.get("is_generator", False):
            return "gen"
        elif func.get("is_arrow", False):
            return "arrow"
        elif self._is_method(func):
            return "method"
        else:
            return "func"

    def _get_method_type(self, method: dict[str, Any]) -> str:
        """Get method type for JavaScript"""
        if method.get("is_constructor", False):
            return "constructor"
        elif method.get("is_getter", False):
            return "getter"
        elif method.get("is_setter", False):
            return "setter"
        elif method.get("is_static", False):
            return "static"
        elif method.get("is_async", False):
            return "async"
        else:
            return "method"

    def _is_method(self, func: dict[str, Any]) -> bool:
        """Check if function is a class method"""
        return func.get("is_method", False) or func.get("class_name") is not None

    def _get_method_class(self, method: dict[str, Any]) -> str:
        """Get the class name for a method"""
        return str(method.get("class_name", "Unknown"))

    def _infer_js_type(self, value: Any) -> str:
        """Infer JavaScript type from value"""
        if value is None:
            return "undefined"

        value_str = str(value).strip()

        # Check for specific patterns
        if value_str == "undefined":
            return "undefined"
        elif value_str == "NaN":
            return "number"  # NaN is a number type in JavaScript
        elif value_str in ["Infinity", "-Infinity"]:
            return "number"  # Infinity is a number type in JavaScript
        elif (
            value_str.startswith('"')
            or value_str.startswith("'")
            or value_str.startswith("`")
        ):
            return "string"
        elif value_str in ["true", "false"]:
            return "boolean"
        elif value_str == "null":
            return "null"
        elif value_str.startswith("[") and value_str.endswith("]"):
            return "array"
        elif value_str.startswith("{") and value_str.endswith("}"):
            return "object"
        elif (
            value_str.startswith("function")
            or value_str.startswith("async function")
            or value_str.startswith("new Function")
            or "=>" in value_str
        ):
            return "function"
        elif value_str.startswith("class"):
            return "class"
        elif value_str.replace(".", "").replace("-", "").isdigit():
            return "number"
        else:
            return "unknown"

    def _determine_scope(self, var: dict[str, Any]) -> str:
        """Determine variable scope"""
        # This would need more context from the parser
        # For now, return basic scope info
        kind = self._get_variable_kind(var)
        if kind == "const" or kind == "let":
            return "block"
        elif kind == "var":
            return "function"
        else:
            return "unknown"

    def _get_variable_kind(self, var: dict[str, Any]) -> str:
        """Get variable declaration kind (const, let, var)"""
        # Check if variable has is_constant flag
        if var.get("is_constant", False):
            return "const"

        # Fall back to parsing raw text
        raw_text = str(var.get("raw_text", "")).strip()
        if raw_text.startswith("const"):
            return "const"
        elif raw_text.startswith("let"):
            return "let"
        elif raw_text.startswith("var"):
            return "var"
        else:
            return "unknown"

    def _get_export_type(self, export: Any) -> str:
        """Get export type"""
        if not isinstance(export, dict):
            return "unknown"
        if export.get("is_default", False):
            return "default"
        elif export.get("is_named", False):
            return "named"
        elif export.get("is_all", False):
            return "all"
        else:
            return "unknown"

    def _get_function_signature(self, func: dict[str, Any]) -> str:
        """Get function signature"""
        name = str(func.get("name", ""))
        params = self._create_full_params(func)
        return_type = func.get("return_type", "")
        if return_type:
            return f"{name}{params} -> {return_type}"
        return f"{name}{params}"

    def _get_class_info(self, cls: dict[str, Any]) -> str:
        """Get class information as formatted string"""
        if cls is None:
            raise TypeError("Cannot format None data")

        if not isinstance(cls, dict):
            raise TypeError(f"Expected dict, got {type(cls)}")

        name = str(cls.get("name", "Unknown"))
        methods = cls.get("methods", [])
        method_count = len(methods) if isinstance(methods, list) else 0

        return f"{name} ({method_count} methods)"

    def _format_json(self, data: dict[str, Any]) -> str:
        """Format data as JSON"""
        import json

        try:
            return json.dumps(data, indent=2, ensure_ascii=False)
        except (TypeError, ValueError) as e:
            return f"# JSON serialization error: {e}\n"

    def format_table(
        self, analysis_result: dict[str, Any], table_type: str = "full"
    ) -> str:
        """Format table output for JavaScript"""
        # Set the format type based on table_type parameter
        original_format_type = self.format_type
        self.format_type = table_type

        try:
            # Use the existing format_structure method
            return self.format_structure(analysis_result)
        finally:
            # Restore original format type
            self.format_type = original_format_type

    def format_summary(self, analysis_result: dict[str, Any]) -> str:
        """Format summary output for JavaScript"""
        return self._format_compact_table(analysis_result)

    def format_structure(self, analysis_result: dict[str, Any]) -> str:
        """Format structure analysis output for JavaScript"""
        return super().format_structure(analysis_result)

    def format_advanced(
        self, analysis_result: dict[str, Any], output_format: str = "json"
    ) -> str:
        """Format advanced analysis output for JavaScript"""
        if output_format == "json":
            return self._format_json(analysis_result)
        elif output_format == "csv":
            return self._format_csv(analysis_result)
        else:
            return self._format_full_table(analysis_result)



#!/usr/bin/env python3
"""
Quick MCP integration test for analyze_code_structure tool
"""

import asyncio
import sys
from pathlib import Path

from tree_sitter_analyzer.mcp.tools.table_format_tool import TableFormatTool

# Add the tree_sitter_analyzer to path
sys.path.insert(0, str(Path(__file__).parent / "tree_sitter_analyzer"))


async def test_mcp_tool() -> None:
    """Test MCP tool integration"""
    test_file = Path("examples/Sample.java")
    if test_file.exists():
        print("Testing MCP tool with Sample.java...")

        # Create tool instance with current directory as project root
        current_dir = str(Path.cwd())
        tool = TableFormatTool(project_root=current_dir)

        # Test full format - use relative path
        result = await tool.execute(
            {"file_path": "examples/Sample.java", "format_type": "full"}
        )
        print("MCP tool test completed successfully")
        print(f"Result length: {len(result.get('table_output', ''))} characters")

        # Test compact format
        result_compact = await tool.execute(
            {"file_path": str(test_file), "format_type": "compact"}
        )
        compact_output = result_compact.get("table_output", "")
        print(f"Compact format length: {len(compact_output)} characters")

        # Test CSV format
        result_csv = await tool.execute(
            {"file_path": str(test_file), "format_type": "csv"}
        )
        print(
            f"CSV format length: {len(result_csv.get('table_output', ''))} characters"
        )

    else:
        print("Sample.java not found, skipping MCP test")


if __name__ == "__main__":
    asyncio.run(test_mcp_tool())



#!/usr/bin/env python3
"""
JavaScript Language Queries

Comprehensive Tree-sitter queries for JavaScript language constructs.
Covers functions, classes, variables, imports, exports, and modern JavaScript features.
Equivalent to Java query coverage for consistent language support.
"""

# JavaScript-specific query library
JAVASCRIPT_QUERIES: dict[str, str] = {
    # --- Basic Structure ---
    "function": """
    (function_declaration) @function
    """,
    "function_declaration": """
    (function_declaration
        name: (identifier) @function_name
        parameters: (formal_parameters) @parameters
        body: (statement_block) @body) @function_declaration
    """,
    "function_expression": """
    (function_expression
        name: (identifier)? @function_name
        parameters: (formal_parameters) @parameters
        body: (statement_block) @body) @function_expression
    """,
    "arrow_function": """
    (arrow_function
        parameters: (formal_parameters) @parameters
        body: (_) @body) @arrow_function
    """,
    "method_definition": """
    (method_definition
        name: (property_identifier) @method_name
        parameters: (formal_parameters) @parameters
        body: (statement_block) @body) @method_definition
    """,
    "async_function": """
    (function_declaration
        "async" @async_keyword
        name: (identifier) @function_name
        parameters: (formal_parameters) @parameters
        body: (statement_block) @body) @async_function
    """,
    "generator_function": """
    (generator_function_declaration
        name: (identifier) @function_name
        parameters: (formal_parameters) @parameters
        body: (statement_block) @body) @generator_function
    """,
    # --- Classes ---
    "class": """
    (class_declaration) @class
    """,
    "class_declaration": """
    (class_declaration
        name: (identifier) @class_name
        (class_heritage)? @superclass
        body: (class_body) @body) @class_declaration
    """,
    "class_expression": """
    (class_expression
        name: (identifier)? @class_name
        (class_heritage)? @superclass
        body: (class_body) @body) @class_expression
    """,
    "class_method": """
    (class_body
        (method_definition
            name: (property_identifier) @method_name
            parameters: (formal_parameters) @parameters
            body: (statement_block) @body)) @class_method
    """,
    "constructor": """
    (method_definition
        name: (property_identifier) @constructor_name
        (#match? @constructor_name "constructor")
        parameters: (formal_parameters) @parameters
        body: (statement_block) @body) @constructor
    """,
    "getter": """
    (method_definition
        "get" @get_keyword
        name: (property_identifier) @getter_name
        body: (statement_block) @body) @getter
    """,
    "setter": """
    (method_definition
        "set" @set_keyword
        name: (property_identifier) @setter_name
        parameters: (formal_parameters) @parameters
        body: (statement_block) @body) @setter
    """,
    "static_method": """
    (method_definition
        "static" @static_keyword
        name: (property_identifier) @method_name
        parameters: (formal_parameters) @parameters
        body: (statement_block) @body) @static_method
    """,
    "private_method": """
    (method_definition
        name: (private_property_identifier) @method_name
        parameters: (formal_parameters) @parameters
        body: (statement_block) @body) @private_method
    """,
    # --- Variables ---
    "variable": """
    (variable_declaration) @variable
    """,
    "var_declaration": """
    (variable_declaration
        (variable_declarator
            name: (identifier) @variable_name
            value: (_)? @value)) @var_declaration
    """,
    "let_declaration": """
    (lexical_declaration
        "let" @let_keyword
        (variable_declarator
            name: (identifier) @variable_name
            value: (_)? @value)) @let_declaration
    """,
    "const_declaration": """
    (lexical_declaration
        "const" @const_keyword
        (variable_declarator
            name: (identifier) @variable_name
            value: (_) @value)) @const_declaration
    """,
    "destructuring_assignment": """
    (variable_declarator
        name: (array_pattern) @array_destructuring
        value: (_) @value) @destructuring_assignment
    """,
    "object_destructuring": """
    (variable_declarator
        name: (object_pattern) @object_destructuring
        value: (_) @value) @object_destructuring
    """,
    # --- Imports and Exports ---
    "import": """
    (import_statement) @import
    """,
    "import_statement": """
    (import_statement
        source: (string) @source) @import_statement
    """,
    "import_default": """
    (import_statement
        (import_clause
            (import_default_specifier
                (identifier) @default_name))
        source: (string) @source) @import_default
    """,
    "import_named": """
    (import_statement
        (import_clause
            (named_imports
                (import_specifier
                    name: (identifier) @import_name
                    alias: (identifier)? @alias)))
        source: (string) @source) @import_named
    """,
    "import_namespace": """
    (import_statement
        (import_clause
            (namespace_import
                (identifier) @namespace_name))
        source: (string) @source) @import_namespace
    """,
    "dynamic_import": """
    (call_expression
        function: (identifier) @import_function
        (#match? @import_function "import")
        arguments: (arguments (string) @source)) @dynamic_import
    """,
    "export": """
    (export_statement) @export
    """,
    "export_default": """
    (export_statement
        "default" @default_keyword
        declaration: (_) @declaration) @export_default
    """,
    "export_named": """
    (export_statement
        (export_clause
            (export_specifier
                name: (identifier) @export_name
                alias: (identifier)? @alias))) @export_named
    """,
    "export_all": """
    (export_statement
        "*" @star
        source: (string) @source) @export_all
    """,
    # --- Objects and Properties ---
    "object": """
    (object) @object
    """,
    "object_literal": """
    (object
        (pair
            key: (_) @key
            value: (_) @value)*) @object_literal
    """,
    "property_definition": """
    (property_definition
        property: (_) @property_name
        value: (_)? @value) @property_definition
    """,
    "computed_property": """
    (pair
        key: (computed_property_name) @computed_key
        value: (_) @value) @computed_property
    """,
    "shorthand_property": """
    (shorthand_property_identifier) @shorthand_property
    """,
    "method_property": """
    (pair
        key: (_) @method_name
        value: (function_expression) @method_function) @method_property
    """,
    # --- Control Flow ---
    "if_statement": """
    (if_statement
        condition: (_) @condition
        consequence: (_) @then_branch
        alternative: (_)? @else_branch) @if_statement
    """,
    "for_statement": """
    (for_statement
        initializer: (_)? @init
        condition: (_)? @condition
        increment: (_)? @update
        body: (_) @body) @for_statement
    """,
    "for_in_statement": """
    (for_in_statement
        left: (_) @variable
        right: (_) @object
        body: (_) @body) @for_in_statement
    """,
    "for_of_statement": """
    (for_of_statement
        left: (_) @variable
        right: (_) @iterable
        body: (_) @body) @for_of_statement
    """,
    "while_statement": """
    (while_statement
        condition: (_) @condition
        body: (_) @body) @while_statement
    """,
    "do_statement": """
    (do_statement
        body: (_) @body
        condition: (_) @condition) @do_statement
    """,
    "switch_statement": """
    (switch_statement
        discriminant: (_) @discriminant
        body: (switch_body) @body) @switch_statement
    """,
    "case_clause": """
    (switch_case
        value: (_) @case_value
        body: (_)* @case_body) @case_clause
    """,
    "default_clause": """
    (switch_default
        body: (_)* @default_body) @default_clause
    """,
    # --- Error Handling ---
    "try_statement": """
    (try_statement
        body: (statement_block) @try_body
        handler: (catch_clause)? @catch_handler
        finalizer: (finally_clause)? @finally_block) @try_statement
    """,
    "catch_clause": """
    (catch_clause
        parameter: (identifier)? @error_parameter
        body: (statement_block) @catch_body) @catch_clause
    """,
    "finally_clause": """
    (finally_clause
        body: (statement_block) @finally_body) @finally_clause
    """,
    "throw_statement": """
    (throw_statement
        argument: (_) @thrown_expression) @throw_statement
    """,
    # --- Modern JavaScript Features ---
    "template_literal": """
    (template_literal) @template_literal
    """,
    "template_substitution": """
    (template_substitution
        expression: (_) @substitution_expr) @template_substitution
    """,
    "spread_element": """
    (spread_element
        argument: (_) @spread_argument) @spread_element
    """,
    "rest_parameter": """
    (rest_parameter
        pattern: (identifier) @rest_name) @rest_parameter
    """,
    "await_expression": """
    (await_expression
        argument: (_) @awaited_expression) @await_expression
    """,
    "yield_expression": """
    (yield_expression
        argument: (_)? @yielded_value) @yield_expression
    """,
    # --- JSX (React) ---
    "jsx_element": """
    (jsx_element
        open_tag: (jsx_opening_element) @open_tag
        close_tag: (jsx_closing_element)? @close_tag) @jsx_element
    """,
    "jsx_self_closing": """
    (jsx_self_closing_element
        name: (_) @element_name) @jsx_self_closing
    """,
    "jsx_attribute": """
    (jsx_attribute
        name: (property_identifier) @attribute_name
        value: (_)? @attribute_value) @jsx_attribute
    """,
    "jsx_expression": """
    (jsx_expression
        expression: (_) @jsx_expression_content) @jsx_expression
    """,
    # --- Comments and Documentation ---
    "comment": """
    (comment) @comment
    """,
    "jsdoc_comment": """
    (comment) @jsdoc_comment
    (#match? @jsdoc_comment "^/\\*\\*")
    """,
    "line_comment": """
    (comment) @line_comment
    (#match? @line_comment "^//")
    """,
    "block_comment": """
    (comment) @block_comment
    (#match? @block_comment "^/\\*(?!\\*)")
    """,
    # --- Framework-specific Patterns ---
    "react_component": """
    (function_declaration
        name: (identifier) @component_name
        (#match? @component_name "^[A-Z]")
        body: (statement_block
            (return_statement
                argument: (jsx_element)))) @react_component
    """,
    "react_hook": """
    (call_expression
        function: (identifier) @hook_name
        (#match? @hook_name "^use[A-Z]")) @react_hook
    """,
    "node_require": """
    (call_expression
        function: (identifier) @require_function
        (#match? @require_function "require")
        arguments: (arguments (string) @module_path)) @node_require
    """,
    "module_exports": """
    (assignment_expression
        left: (member_expression
            object: (identifier) @module_object
            (#match? @module_object "module")
            property: (property_identifier) @exports_property
            (#match? @exports_property "exports"))
        right: (_) @exported_value) @module_exports
    """,
    # --- Name-only Extraction ---
    "function_name": """
    (function_declaration
        name: (identifier) @function_name)
    """,
    "class_name": """
    (class_declaration
        name: (identifier) @class_name)
    """,
    "variable_name": """
    (variable_declarator
        name: (identifier) @variable_name)
    """,
    # --- Advanced Patterns ---
    "closure": """
    (function_expression
        body: (statement_block
            (return_statement
                argument: (function_expression)))) @closure
    """,
    "callback_function": """
    (call_expression
        arguments: (arguments
            (function_expression) @callback_function)) @callback_call
    """,
    "promise_chain": """
    (call_expression
        function: (member_expression
            object: (_) @promise_object
            property: (property_identifier) @chain_method
            (#match? @chain_method "^(then|catch|finally)$"))) @promise_chain
    """,
    "event_listener": """
    (call_expression
        function: (member_expression
            property: (property_identifier) @listener_method
            (#match? @listener_method "^(addEventListener|on)$"))
        arguments: (arguments
            (string) @event_type
            (function_expression) @event_handler)) @event_listener
    """,
    "iife": """
    (call_expression
        function: (function_expression) @iife_function) @iife
    """,
    "module_pattern": """
    (variable_declarator
        name: (identifier) @module_name
        value: (call_expression
            function: (function_expression) @module_function)) @module_pattern
    """,
}

# Query descriptions
JAVASCRIPT_QUERY_DESCRIPTIONS: dict[str, str] = {
    "function": "Search JavaScript function declarations",
    "function_declaration": "Search function declarations with details",
    "function_expression": "Search function expressions",
    "arrow_function": "Search arrow functions",
    "method_definition": "Search method definitions",
    "async_function": "Search async function declarations",
    "generator_function": "Search generator functions",
    "class": "Search JavaScript class declarations",
    "class_declaration": "Search class declarations with details",
    "class_expression": "Search class expressions",
    "class_method": "Search class methods",
    "constructor": "Search class constructors",
    "getter": "Search getter methods",
    "setter": "Search setter methods",
    "static_method": "Search static methods",
    "private_method": "Search private methods",
    "variable": "Search variable declarations",
    "var_declaration": "Search var declarations",
    "let_declaration": "Search let declarations",
    "const_declaration": "Search const declarations",
    "destructuring_assignment": "Search destructuring assignments",
    "object_destructuring": "Search object destructuring",
    "import": "Search import statements",
    "import_statement": "Search import statements with details",
    "import_default": "Search default imports",
    "import_named": "Search named imports",
    "import_namespace": "Search namespace imports",
    "dynamic_import": "Search dynamic imports",
    "export": "Search export statements",
    "export_default": "Search default exports",
    "export_named": "Search named exports",
    "export_all": "Search export all statements",
    "object": "Search object literals",
    "object_literal": "Search object literals with properties",
    "property_definition": "Search property definitions",
    "computed_property": "Search computed properties",
    "shorthand_property": "Search shorthand properties",
    "method_property": "Search method properties",
    "if_statement": "Search if statements",
    "for_statement": "Search for loops",
    "for_in_statement": "Search for-in loops",
    "for_of_statement": "Search for-of loops",
    "while_statement": "Search while loops",
    "do_statement": "Search do-while loops",
    "switch_statement": "Search switch statements",
    "case_clause": "Search switch case clauses",
    "default_clause": "Search switch default clauses",
    "try_statement": "Search try-catch statements",
    "catch_clause": "Search catch clauses",
    "finally_clause": "Search finally clauses",
    "throw_statement": "Search throw statements",
    "template_literal": "Search template literals",
    "template_substitution": "Search template substitutions",
    "spread_element": "Search spread elements",
    "rest_parameter": "Search rest parameters",
    "await_expression": "Search await expressions",
    "yield_expression": "Search yield expressions",
    "jsx_element": "Search JSX elements",
    "jsx_self_closing": "Search self-closing JSX elements",
    "jsx_attribute": "Search JSX attributes",
    "jsx_expression": "Search JSX expressions",
    "comment": "Search all comments",
    "jsdoc_comment": "Search JSDoc comments",
    "line_comment": "Search line comments",
    "block_comment": "Search block comments",
    "react_component": "Search React components",
    "react_hook": "Search React hooks",
    "node_require": "Search Node.js require statements",
    "module_exports": "Search Node.js module exports",
    "function_name": "Search function names only",
    "class_name": "Search class names only",
    "variable_name": "Search variable names only",
    "closure": "Search closures",
    "callback_function": "Search callback functions passed as arguments",
    "promise_chain": "Search Promise chain patterns (.then, .catch, .finally)",
    "event_listener": "Search event listener registrations",
    "iife": "Search immediately invoked function expressions",
    "module_pattern": "Search module patterns",
}

# Legacy query definitions for backward compatibility
FUNCTIONS = """
(function_declaration
    name: (identifier) @function.name
    parameters: (formal_parameters) @function.params
    body: (statement_block) @function.body) @function.declaration

(function_expression
    name: (identifier)? @function.name
    parameters: (formal_parameters) @function.params
    body: (statement_block) @function.body) @function.expression

(arrow_function
    parameters: (formal_parameters) @function.params
    body: (_) @function.body) @function.arrow

(method_definition
    name: (property_identifier) @function.name
    parameters: (formal_parameters) @function.params
    body: (statement_block) @function.body) @method.definition
"""

CLASSES = """
(class_declaration
    name: (identifier) @class.name
    (class_heritage)? @class.superclass
    body: (class_body) @class.body) @class.declaration
"""

VARIABLES = """
(variable_declaration
    (variable_declarator
        name: (identifier) @variable.name
        value: (_)? @variable.value)) @variable.declaration

(lexical_declaration
    (variable_declarator
        name: (identifier) @variable.name
        value: (_)? @variable.value)) @variable.lexical
"""

IMPORTS = """
(import_statement
    source: (string) @import.source) @import.statement

(import_statement
    (import_clause
        (named_imports
            (import_specifier
                name: (identifier) @import.name
                alias: (identifier)? @import.alias)))) @import.named

(import_statement
    (import_clause
        (import_default_specifier
            (identifier) @import.default))) @import.default

(import_statement
    (import_clause
        (namespace_import
            (identifier) @import.namespace))) @import.namespace
"""

EXPORTS = """
(export_statement
    declaration: (_) @export.declaration) @export.statement

(export_statement
    (export_clause
        (export_specifier
            name: (identifier) @export.name
            alias: (identifier)? @export.alias))) @export.named
"""

OBJECTS = """
(object
    (pair
        key: (_) @property.key
        value: (_) @property.value)) @object.literal

(property_definition
    property: (_) @property.name
    value: (_)? @property.value) @property.definition
"""

COMMENTS = """
(comment) @comment
"""

# Convert to ALL_QUERIES format for dynamic loader compatibility
ALL_QUERIES = {}
for query_name, query_string in JAVASCRIPT_QUERIES.items():
    description = JAVASCRIPT_QUERY_DESCRIPTIONS.get(query_name, "No description")
    ALL_QUERIES[query_name] = {"query": query_string, "description": description}

# Add legacy queries for backward compatibility
ALL_QUERIES["functions"] = {
    "query": FUNCTIONS,
    "description": "Search all function declarations, expressions, and methods",
}
ALL_QUERIES["classes"] = {
    "query": CLASSES,
    "description": "Search all class declarations and expressions",
}
ALL_QUERIES["variables"] = {
    "query": VARIABLES,
    "description": "Search all variable declarations (var, let, const)",
}
ALL_QUERIES["imports"] = {
    "query": IMPORTS,
    "description": "Search all import statements and clauses",
}
ALL_QUERIES["exports"] = {
    "query": EXPORTS,
    "description": "Search all export statements",
}
ALL_QUERIES["objects"] = {
    "query": OBJECTS,
    "description": "Search object literals and property definitions",
}
ALL_QUERIES["comments"] = {"query": COMMENTS, "description": "Search all comments"}

# Add missing method queries
ALL_QUERIES["method"] = {
    "query": """
(method_definition
    name: (property_identifier) @method_name
    parameters: (formal_parameters) @parameters
    body: (statement_block) @body) @method_definition
""",
    "description": "Search method definitions",
}
ALL_QUERIES["methods"] = {
    "query": """
(method_definition
    name: (property_identifier) @method_name
    parameters: (formal_parameters) @parameters
    body: (statement_block) @body) @method_definition
""",
    "description": "Search method definitions",
}


def get_javascript_query(name: str) -> str:
    """
    Get the specified JavaScript query

    Args:
        name: Query name

    Returns:
        Query string

    Raises:
        ValueError: When query is not found
    """
    if name not in JAVASCRIPT_QUERIES:
        available = list(JAVASCRIPT_QUERIES.keys())
        raise ValueError(
            f"JavaScript query '{name}' does not exist. Available: {available}"
        )

    return JAVASCRIPT_QUERIES[name]


def get_javascript_query_description(name: str) -> str:
    """
    Get the description of the specified JavaScript query

    Args:
        name: Query name

    Returns:
        Query description
    """
    return JAVASCRIPT_QUERY_DESCRIPTIONS.get(name, "No description")


def get_query(name: str) -> str:
    """Get a specific query by name."""
    if name in ALL_QUERIES:
        return ALL_QUERIES[name]["query"]
    raise ValueError(
        f"Query '{name}' not found. Available queries: {list(ALL_QUERIES.keys())}"
    )


def get_all_queries() -> dict:
    """Get all available queries."""
    return ALL_QUERIES


def list_queries() -> list:
    """List all available query names."""
    return list(ALL_QUERIES.keys())


def get_available_javascript_queries() -> list[str]:
    """
    Get list of available JavaScript queries

    Returns:
        List of query names
    """
    return list(JAVASCRIPT_QUERIES.keys())



#!/usr/bin/env python3
"""
Error Handler for MCP Server

This module provides comprehensive error handling and recovery
mechanisms for the MCP server operations.
"""

import asyncio
import logging
import traceback
from collections.abc import Callable
from datetime import datetime
from enum import Enum
from functools import wraps
from typing import Any

logger = logging.getLogger(__name__)


class ErrorSeverity(Enum):
    """Error severity levels"""

    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"
    CRITICAL = "critical"


class ErrorCategory(Enum):
    """Error categories for classification"""

    FILE_ACCESS = "file_access"
    PARSING = "parsing"
    ANALYSIS = "analysis"
    NETWORK = "network"
    VALIDATION = "validation"
    RESOURCE = "resource"
    CONFIGURATION = "configuration"
    UNKNOWN = "unknown"


class MCPError(Exception):
    """Base exception class for MCP-specific errors"""

    def __init__(
        self,
        message: str,
        category: ErrorCategory = ErrorCategory.UNKNOWN,
        severity: ErrorSeverity = ErrorSeverity.MEDIUM,
        details: dict[str, Any] | None = None,
        recoverable: bool = True,
    ):
        super().__init__(message)
        self.message = message
        self.category = category
        self.severity = severity
        self.details = details or {}
        self.recoverable = recoverable
        self.timestamp = datetime.now()

    def to_dict(self) -> dict[str, Any]:
        """Convert error to dictionary representation"""
        return {
            "error_type": self.__class__.__name__,
            "message": self.message,
            "category": self.category.value,
            "severity": self.severity.value,
            "details": self.details,
            "recoverable": self.recoverable,
            "timestamp": self.timestamp.isoformat(),
        }


class FileAccessError(MCPError):
    """Error related to file access operations"""

    def __init__(self, message: str, file_path: str, **kwargs: Any):
        super().__init__(
            message,
            category=ErrorCategory.FILE_ACCESS,
            details={"file_path": file_path},
            **kwargs,
        )


class ParsingError(MCPError):
    """Error related to code parsing operations"""

    def __init__(
        self,
        message: str,
        file_path: str,
        language: str | None = None,
        **kwargs: Any,
    ):
        super().__init__(
            message,
            category=ErrorCategory.PARSING,
            details={"file_path": file_path, "language": language},
            **kwargs,
        )


class AnalysisError(MCPError):
    """Error related to code analysis operations"""

    def __init__(self, message: str, operation: str, **kwargs: Any):
        super().__init__(
            message,
            category=ErrorCategory.ANALYSIS,
            details={"operation": operation},
            **kwargs,
        )


class ValidationError(MCPError):
    """Error related to input validation"""

    def __init__(self, message: str, field: str, value: Any = None, **kwargs: Any):
        super().__init__(
            message,
            category=ErrorCategory.VALIDATION,
            details={
                "field": field,
                "value": str(value) if value is not None else None,
            },
            **kwargs,
        )


class ResourceError(MCPError):
    """Error related to resource operations"""

    def __init__(self, message: str, resource_uri: str, **kwargs: Any):
        super().__init__(
            message,
            category=ErrorCategory.RESOURCE,
            details={"resource_uri": resource_uri},
            **kwargs,
        )


class ErrorHandler:
    """
    Centralized error handling and recovery system

    Provides error classification, logging, recovery mechanisms,
    and error statistics for the MCP server.
    """

    def __init__(self) -> None:
        """Initialize error handler"""
        self.error_counts: dict[str, int] = {}
        self.error_history: list[dict[str, Any]] = []
        self.max_history_size = 1000
        self.recovery_strategies: dict[type[Exception], Callable] = {}

        # Register default recovery strategies
        self._register_default_strategies()

        logger.info("Error handler initialized")

    def _register_default_strategies(self) -> None:
        """Register default error recovery strategies"""

        def file_not_found_recovery(
            error: FileNotFoundError, context: dict[str, Any]
        ) -> dict[str, Any]:
            """Recovery strategy for file not found errors"""
            return {
                "error": f"File not found: {context.get('file_path', 'unknown')}",
                "suggestion": "Please check the file path and ensure the file exists",
                "recoverable": True,
            }

        def permission_error_recovery(
            error: PermissionError, context: dict[str, Any]
        ) -> dict[str, Any]:
            """Recovery strategy for permission errors"""
            return {
                "error": f"Permission denied: {context.get('file_path', 'unknown')}",
                "suggestion": "Please check file permissions or run with appropriate privileges",
                "recoverable": False,
            }

        def value_error_recovery(
            error: ValueError, context: dict[str, Any]
        ) -> dict[str, Any]:
            """Recovery strategy for value errors"""
            return {
                "error": f"Invalid value: {str(error)}",
                "suggestion": "Please check input parameters and try again",
                "recoverable": True,
            }

        self.recovery_strategies.update(
            {
                FileNotFoundError: file_not_found_recovery,
                PermissionError: permission_error_recovery,
                ValueError: value_error_recovery,
            }
        )

    def register_recovery_strategy(
        self,
        exception_type: type[Exception],
        strategy: Callable[[Exception, dict[str, Any]], dict[str, Any]],
    ) -> None:
        """
        Register a custom recovery strategy for an exception type

        Args:
            exception_type: Type of exception to handle
            strategy: Recovery function that takes (exception, context) and returns recovery info
        """
        self.recovery_strategies[exception_type] = strategy
        logger.debug(f"Registered recovery strategy for {exception_type.__name__}")

    def handle_error(
        self,
        error: Exception,
        context: dict[str, Any] | None = None,
        operation: str = "unknown",
    ) -> dict[str, Any]:
        """
        Handle an error with classification, logging, and recovery

        Args:
            error: The exception that occurred
            context: Additional context information
            operation: Name of the operation that failed

        Returns:
            Error information dictionary with recovery suggestions
        """
        context = context or {}

        # Classify error
        if isinstance(error, MCPError):
            error_info = error.to_dict()
        else:
            error_info = self._classify_error(error, context, operation)

        # Log error
        self._log_error(error, error_info, context, operation)

        # Update statistics
        self._update_error_stats(error_info)

        # Add to history
        self._add_to_history(error_info, context, operation)

        # Attempt recovery
        recovery_info = self._attempt_recovery(error, context)
        if recovery_info:
            error_info.update(recovery_info)

        return error_info

    def _classify_error(
        self, error: Exception, context: dict[str, Any], operation: str
    ) -> dict[str, Any]:
        """
        Classify a generic exception into MCP error categories

        Args:
            error: The exception to classify
            context: Error context
            operation: Operation that failed

        Returns:
            Error information dictionary
        """
        error_type = type(error).__name__
        message = str(error)

        # Determine category based on error type and context
        category = ErrorCategory.UNKNOWN
        severity = ErrorSeverity.MEDIUM
        recoverable = True

        if isinstance(
            error, FileNotFoundError | IsADirectoryError | NotADirectoryError
        ):
            category = ErrorCategory.FILE_ACCESS
            severity = ErrorSeverity.MEDIUM
        elif isinstance(error, PermissionError):
            category = ErrorCategory.FILE_ACCESS
            severity = ErrorSeverity.HIGH
            recoverable = False
        elif isinstance(error, ValueError | TypeError):
            category = ErrorCategory.VALIDATION
            severity = ErrorSeverity.LOW
        elif isinstance(error, OSError | IOError):
            category = ErrorCategory.FILE_ACCESS
            severity = ErrorSeverity.HIGH
        elif isinstance(error, RuntimeError | AttributeError):
            category = ErrorCategory.ANALYSIS
            severity = ErrorSeverity.MEDIUM
        elif isinstance(error, MemoryError):
            category = ErrorCategory.RESOURCE
            severity = ErrorSeverity.CRITICAL
            recoverable = False
        elif isinstance(error, asyncio.TimeoutError):
            category = ErrorCategory.NETWORK
            severity = ErrorSeverity.MEDIUM

        return {
            "error_type": error_type,
            "message": message,
            "category": category.value,
            "severity": severity.value,
            "details": context,
            "recoverable": recoverable,
            "timestamp": datetime.now().isoformat(),
            "operation": operation,
            "traceback": traceback.format_exc(),
        }

    def _log_error(
        self,
        error: Exception,
        error_info: dict[str, Any],
        context: dict[str, Any],
        operation: str,
    ) -> None:
        """
        Log error with appropriate level based on severity

        Args:
            error: The original exception
            error_info: Classified error information
            context: Error context
            operation: Operation that failed
        """
        severity = error_info.get("severity", "medium")
        message = f"Error in {operation}: {error_info['message']}"

        if severity == "critical":
            logger.critical(
                message, extra={"error_info": error_info, "context": context}
            )
        elif severity == "high":
            logger.error(message, extra={"error_info": error_info, "context": context})
        elif severity == "medium":
            logger.warning(
                message, extra={"error_info": error_info, "context": context}
            )
        else:
            logger.info(message, extra={"error_info": error_info, "context": context})

    def _update_error_stats(self, error_info: dict[str, Any]) -> None:
        """
        Update error statistics

        Args:
            error_info: Error information
        """
        error_type = error_info.get("error_type", "Unknown")
        category = error_info.get("category", "unknown")

        # Count by type
        self.error_counts[f"type:{error_type}"] = (
            self.error_counts.get(f"type:{error_type}", 0) + 1
        )

        # Count by category
        self.error_counts[f"category:{category}"] = (
            self.error_counts.get(f"category:{category}", 0) + 1
        )

        # Count by severity
        severity = error_info.get("severity", "medium")
        self.error_counts[f"severity:{severity}"] = (
            self.error_counts.get(f"severity:{severity}", 0) + 1
        )

    def _add_to_history(
        self, error_info: dict[str, Any], context: dict[str, Any], operation: str
    ) -> None:
        """
        Add error to history with size limit

        Args:
            error_info: Error information
            context: Error context
            operation: Operation that failed
        """
        history_entry = {**error_info, "context": context, "operation": operation}

        self.error_history.append(history_entry)

        # Maintain history size limit
        if len(self.error_history) > self.max_history_size:
            self.error_history = self.error_history[-self.max_history_size :]

    def _attempt_recovery(
        self, error: Exception, context: dict[str, Any]
    ) -> dict[str, Any] | None:
        """
        Attempt to recover from error using registered strategies

        Args:
            error: The exception to recover from
            context: Error context

        Returns:
            Recovery information or None
        """
        error_type = type(error)

        # Try exact type match first
        if error_type in self.recovery_strategies:
            try:
                result = self.recovery_strategies[error_type](error, context)
                return result if result is not None else {}
            except Exception as recovery_error:
                logger.warning(f"Recovery strategy failed: {recovery_error}")

        # Try parent class matches
        for registered_type, strategy in self.recovery_strategies.items():
            if isinstance(error, registered_type):
                try:
                    result = strategy(error, context)
                    return result if result is not None else {}
                except Exception as recovery_error:
                    logger.warning(f"Recovery strategy failed: {recovery_error}")

        return None

    def get_error_stats(self) -> dict[str, Any]:
        """
        Get error statistics

        Returns:
            Dictionary containing error statistics
        """
        total_errors = (
            sum(self.error_counts.values()) // 3
        )  # Divide by 3 because we count type, category, severity

        return {
            "total_errors": total_errors,
            "error_counts": self.error_counts.copy(),
            "recent_errors": len(
                [
                    e
                    for e in self.error_history
                    if (datetime.now() - datetime.fromisoformat(e["timestamp"])).seconds
                    < 3600
                ]
            ),
            "history_size": len(self.error_history),
        }

    def get_recent_errors(self, limit: int = 10) -> list[dict[str, Any]]:
        """
        Get recent errors from history

        Args:
            limit: Maximum number of errors to return

        Returns:
            List of recent error entries
        """
        return self.error_history[-limit:] if self.error_history else []

    def clear_history(self) -> None:
        """Clear error history and reset statistics"""
        self.error_history.clear()
        self.error_counts.clear()
        logger.info("Error history and statistics cleared")


def handle_mcp_errors(
    operation: str = "unknown",
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """
    Decorator for automatic error handling in MCP operations

    Args:
        operation: Name of the operation for logging

    Returns:
        Decorated function with error handling
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
            try:
                return await func(*args, **kwargs)
            except RuntimeError as e:
                # Handle initialization errors specifically
                if "not fully initialized" in str(e):
                    logger.warning(
                        f"Request received before initialization complete: {operation}"
                    )
                    raise MCPError(
                        "Server is still initializing. Please wait a moment and try again.",
                        category=ErrorCategory.CONFIGURATION,
                        severity=ErrorSeverity.LOW,
                    ) from e
                # Handle other runtime errors normally
                error_handler = get_error_handler()
                context = {
                    "function": func.__name__,
                    "args": str(args)[:200],  # Limit length
                    "kwargs": str(kwargs)[:200],
                }
                error_info = error_handler.handle_error(e, context, operation)
                raise
            except Exception as e:
                error_handler = get_error_handler()
                context = {
                    "function": func.__name__,
                    "args": str(args)[:200],  # Limit length
                    "kwargs": str(kwargs)[:200],
                }
                error_info = error_handler.handle_error(e, context, operation)

                # Re-raise as MCPError if not already
                if not isinstance(e, MCPError):
                    raise AnalysisError(
                        f"Operation failed: {error_info['message']}",
                        operation=operation,
                        severity=ErrorSeverity(error_info["severity"]),
                    ) from e
                raise

        @wraps(func)
        def sync_wrapper(*args: Any, **kwargs: Any) -> Any:
            try:
                return func(*args, **kwargs)
            except Exception as e:
                error_handler = get_error_handler()
                context = {
                    "function": func.__name__,
                    "args": str(args)[:200],
                    "kwargs": str(kwargs)[:200],
                }
                error_info = error_handler.handle_error(e, context, operation)

                if not isinstance(e, MCPError):
                    raise AnalysisError(
                        f"Operation failed: {error_info['message']}",
                        operation=operation,
                        severity=ErrorSeverity(error_info["severity"]),
                    ) from e
                raise

        return async_wrapper if asyncio.iscoroutinefunction(func) else sync_wrapper

    return decorator


# Global error handler instance
_error_handler = ErrorHandler()


def get_error_handler() -> ErrorHandler:
    """
    Get the global error handler instance

    Returns:
        Global error handler
    """
    return _error_handler



#!/usr/bin/env python3
"""
Rust Language Plugin

Provides Rust-specific parsing and element extraction functionality.
"""

import re
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import tree_sitter

    from ..core.analysis_engine import AnalysisRequest
    from ..models import AnalysisResult

from ..encoding_utils import extract_text_slice, safe_encode
from ..models import Class, Function, Import, Variable
from ..plugins.base import ElementExtractor, LanguagePlugin
from ..utils import log_debug, log_error


class RustElementExtractor(ElementExtractor):
    """Rust-specific element extractor"""

    def __init__(self) -> None:
        """Initialize the Rust element extractor."""
        self.current_module: str = ""
        self.current_file: str = ""
        self.source_code: str = ""
        self.content_lines: list[str] = []
        self._node_text_cache: dict[tuple[int, int], str] = {}
        self.impl_blocks: list[dict[str, Any]] = []
        self.modules: list[dict[str, Any]] = []

    def extract_functions(
        self, tree: "tree_sitter.Tree", source_code: str
    ) -> list[Function]:
        """Extract Rust function declarations"""
        self.source_code = source_code
        self.content_lines = source_code.split("\n")
        self._reset_caches()

        functions: list[Function] = []

        # Use tree traversal to find function_item
        self._traverse_and_extract(
            tree.root_node,
            {"function_item": self._extract_function},
            functions,
        )

        log_debug(f"Extracted {len(functions)} Rust functions")
        return functions

    def extract_classes(
        self, tree: "tree_sitter.Tree", source_code: str
    ) -> list[Class]:
        """Extract Rust struct, enum, trait, and impl definitions"""
        self.source_code = source_code
        self.content_lines = source_code.split("\n")
        self._reset_caches()

        # Extract modules first
        self._extract_modules(tree.root_node)

        classes: list[Class] = []

        extractors = {
            "struct_item": self._extract_struct,
            "enum_item": self._extract_enum,
            "trait_item": self._extract_trait,
            "impl_item": self._extract_impl,  # Impl blocks are treated as related to classes
        }

        self._traverse_and_extract(
            tree.root_node,
            extractors,
            classes,
        )

        # Process collected impl blocks and add them to classes list if they are standalone
        # Or we might want to return them as separate metadata.
        # For now, we'll include impl blocks as Class objects with type='impl' for visibility
        for _impl in self.impl_blocks:
            # Creating a Class object for impl block to represent it in the structure
            pass

        log_debug(f"Extracted {len(classes)} Rust structs/enums/traits")
        return classes

    def extract_variables(
        self, tree: "tree_sitter.Tree", source_code: str
    ) -> list[Variable]:
        """Extract Rust struct fields"""
        self.source_code = source_code
        self.content_lines = source_code.split("\n")
        self._reset_caches()

        variables: list[Variable] = []

        # We extract fields from struct definitions
        extractors = {
            "field_declaration": self._extract_field,
        }

        self._traverse_and_extract(
            tree.root_node,
            extractors,
            variables,
        )

        log_debug(f"Extracted {len(variables)} Rust fields")
        return variables

    def extract_imports(
        self, tree: "tree_sitter.Tree", source_code: str
    ) -> list[Import]:
        """Extract Rust use declarations"""
        self.source_code = source_code
        self.content_lines = source_code.split("\n")
        self._reset_caches()

        imports: list[Import] = []

        # We extract use declarations
        extractors = {
            "use_declaration": self._extract_import,
        }

        self._traverse_and_extract(
            tree.root_node,
            extractors,
            imports,
        )

        log_debug(f"Extracted {len(imports)} Rust imports")
        return imports

    def _extract_import(self, node: "tree_sitter.Node") -> Import | None:
        """Extract import statement (use declaration)"""
        try:
            raw_text = self._get_node_text(node)
            start_line = node.start_point[0] + 1
            end_line = node.end_point[0] + 1

            # Extract name (the path)
            # use std::collections::HashMap;
            # The actual path is within the node children.
            # Typically we can just use raw text or try to parse it.
            # For simplicity, we'll use raw text as the import statement.

            return Import(
                name=raw_text,  # Use full statement as name for now, or parse better
                start_line=start_line,
                end_line=end_line,
                raw_text=raw_text,
                language="rust",
                import_statement=raw_text,
            )
        except Exception as e:
            log_error(f"Error extracting Rust import: {e}")
            return None

    def _reset_caches(self) -> None:
        """Reset performance caches"""
        self._node_text_cache.clear()
        # Modules and impls persist across extraction calls within the same file analysis
        # but we clear them here if we assume sequential full extraction calls.
        # Ideally, we should call extract_modules separately or share state.
        # For simplicity in this architecture, we might re-extract or just check if empty.
        if not self.source_code:
            self.modules.clear()
            self.impl_blocks.clear()

    def _traverse_and_extract(
        self,
        node: "tree_sitter.Node",
        extractors: dict[str, Any],
        results: list[Any],
    ) -> None:
        """Recursive traversal to find and extract elements"""
        if node.type in extractors:
            element = extractors[node.type](node)
            if element:
                results.append(element)

        for child in node.children:
            self._traverse_and_extract(child, extractors, results)

    def _extract_modules(self, node: "tree_sitter.Node") -> None:
        """Extract module information"""
        if node.type == "mod_item":
            self._extract_module(node)

        for child in node.children:
            self._extract_modules(child)

    def _extract_module(self, node: "tree_sitter.Node") -> None:
        """Extract single module"""
        try:
            name_node = node.child_by_field_name("name")
            if name_node:
                name = self._get_node_text(name_node)
                visibility = self._extract_visibility(node)

                self.modules.append(
                    {
                        "name": name,
                        "visibility": visibility,
                        "line_range": {
                            "start": node.start_point[0] + 1,
                            "end": node.end_point[0] + 1,
                        },
                    }
                )
        except Exception as e:
            log_error(f"Error extracting module: {e}")

    def _extract_function(self, node: "tree_sitter.Node") -> Function | None:
        """Extract function information"""
        try:
            name_node = node.child_by_field_name("name")
            if not name_node:
                return None

            name = self._get_node_text(name_node)
            start_line = node.start_point[0] + 1
            end_line = node.end_point[0] + 1

            # Parameters
            parameters = []
            params_node = node.child_by_field_name("parameters")
            if params_node:
                for child in params_node.children:
                    if child.type == "parameter":
                        parameters.append(self._get_node_text(child))
                    elif child.type == "self_parameter":
                        parameters.append("self")

            # Return type
            return_type = "()"
            ret_node = node.child_by_field_name("return_type")
            if ret_node:
                return_type = self._get_node_text(ret_node)
                # Remove "->" prefix if captured
                if return_type.startswith("->"):
                    return_type = return_type[2:].strip()

            # Visibility
            visibility = self._extract_visibility(node)

            # Async - check function_modifiers node for async keyword
            is_async = False
            for child in node.children:
                if child.type == "function_modifiers":
                    # Check if async is in the modifiers
                    for modifier in child.children:
                        if modifier.type == "async":
                            is_async = True
                            break
                    if is_async:
                        break
                # Also check for direct async child (older tree-sitter versions)
                elif child.type == "async":
                    is_async = True
                    break

            # Docstring
            docstring = self._extract_docstring(node)

            # Raw text
            raw_text = self._get_node_text(node)

            # Add to Function object
            # Note: We're using dynamic attributes for Rust-specific fields
            func = Function(
                name=name,
                start_line=start_line,
                end_line=end_line,
                raw_text=raw_text,
                language="rust",
                parameters=parameters,
                return_type=return_type,
                visibility=visibility,
                docstring=docstring,
            )
            # Attach Rust-specific attributes
            func.is_async = is_async

            return func

        except Exception as e:
            log_error(f"Error extracting Rust function: {e}")
            return None

    def _extract_struct(self, node: "tree_sitter.Node") -> Class | None:
        """Extract struct information"""
        return self._extract_type_def(node, "struct")

    def _extract_enum(self, node: "tree_sitter.Node") -> Class | None:
        """Extract enum information"""
        return self._extract_type_def(node, "enum")

    def _extract_trait(self, node: "tree_sitter.Node") -> Class | None:
        """Extract trait information"""
        return self._extract_type_def(node, "trait")

    def _extract_type_def(
        self, node: "tree_sitter.Node", type_name: str
    ) -> Class | None:
        """Generic type definition extractor"""
        try:
            name_node = node.child_by_field_name("name")
            if not name_node:
                return None

            name = self._get_node_text(name_node)
            start_line = node.start_point[0] + 1
            end_line = node.end_point[0] + 1
            visibility = self._extract_visibility(node)

            raw_text = self._get_node_text(node)

            cls = Class(
                name=name,
                start_line=start_line,
                end_line=end_line,
                raw_text=raw_text,
                language="rust",
                class_type=type_name,
                visibility=visibility,
            )

            # Extract implemented traits (for structs/enums) or supertraits (for traits)
            # This is complex in Rust as impls are separate.
            # We can scan for derive macros here.
            derives = self._extract_derives(node)
            if derives:
                # Add derives to implemented interfaces list for display
                cls.implements_interfaces = derives

            return cls

        except Exception as e:
            log_error(f"Error extracting Rust {type_name}: {e}")
            return None

    def _extract_impl(self, node: "tree_sitter.Node") -> None:
        """Extract impl block information"""
        try:
            trait_node = node.child_by_field_name("trait")
            type_node = node.child_by_field_name("type")

            trait_name = self._get_node_text(trait_node) if trait_node else None
            type_name = self._get_node_text(type_node) if type_node else None

            if type_name:
                self.impl_blocks.append(
                    {
                        "type": type_name,
                        "trait": trait_name,
                        "line_range": {
                            "start": node.start_point[0] + 1,
                            "end": node.end_point[0] + 1,
                        },
                    }
                )
        except Exception as e:
            log_error(f"Error extracting impl block: {e}")

    def _extract_field(self, node: "tree_sitter.Node") -> Variable | None:
        """Extract struct field"""
        try:
            name_node = node.child_by_field_name("name")
            type_node = node.child_by_field_name("type")

            if not name_node or not type_node:
                return None

            name = self._get_node_text(name_node)
            field_type = self._get_node_text(type_node)
            start_line = node.start_point[0] + 1
            end_line = node.end_point[0] + 1
            visibility = self._extract_visibility(node)

            raw_text = self._get_node_text(node)
            docstring = self._extract_docstring(node)

            return Variable(
                name=name,
                start_line=start_line,
                end_line=end_line,
                raw_text=raw_text,
                language="rust",
                variable_type=field_type,
                visibility=visibility,
                docstring=docstring,
            )
        except Exception as e:
            log_error(f"Error extracting Rust field: {e}")
            return None

    def _extract_visibility(self, node: "tree_sitter.Node") -> str:
        """Extract visibility modifier"""
        for child in node.children:
            if child.type == "visibility_modifier":
                return self._get_node_text(child)
        return "private"  # Default in Rust

    def _extract_docstring(self, node: "tree_sitter.Node") -> str | None:
        """Extract doc comments (/// or /** ... */)"""
        # In tree-sitter-rust, doc comments are often 'line_comment' or 'block_comment'
        # preceding the item, or attributes.
        # But often they are just comments attached to the node if we look at previous siblings
        # or they are part of the node as 'outer attributes' which are 'attribute_item'

        docs = []
        # Look for attribute items that are doc comments
        # This simple implementation might need refinement based on actual tree structure
        for child in node.children:
            if child.type == "line_comment" and self._get_node_text(child).startswith(
                "///"
            ):
                docs.append(self._get_node_text(child)[3:].strip())
            elif child.type == "block_comment" and self._get_node_text(
                child
            ).startswith("/**"):
                content = self._get_node_text(child)
                # Strip /** and */
                if content.startswith("/**") and content.endswith("*/"):
                    docs.append(content[3:-2].strip())

        if docs:
            return "\n".join(docs)

        # Fallback: check source lines before start_line (similar to Java)
        start_line = node.start_point[0]
        if start_line > 0:
            # Check previous lines
            pass

        return None

    def _extract_derives(self, node: "tree_sitter.Node") -> list[str]:
        """Extract derived traits from attributes"""
        derives = []
        for child in node.children:
            if child.type == "attribute_item":
                text = self._get_node_text(child)
                if "derive" in text:
                    # Naive parsing of #[derive(Debug, Clone)]
                    match = re.search(r"derive\((.*?)\)", text)
                    if match:
                        traits = match.group(1).split(",")
                        derives.extend([t.strip() for t in traits if t.strip()])
        return derives

    def _get_node_text(self, node: "tree_sitter.Node") -> str:
        """Get node text with caching using position-based keys"""
        cache_key = (node.start_byte, node.end_byte)
        if cache_key in self._node_text_cache:
            return self._node_text_cache[cache_key]

        try:
            start_byte = node.start_byte
            end_byte = node.end_byte
            encoding = "utf-8"  # Default
            content_bytes = safe_encode("\n".join(self.content_lines), encoding)
            text = extract_text_slice(content_bytes, start_byte, end_byte, encoding)
            self._node_text_cache[cache_key] = text
            return text
        except Exception:
            return ""


class RustPlugin(LanguagePlugin):
    """Rust language plugin implementation"""

    def __init__(self) -> None:
        """Initialize the Rust language plugin."""
        super().__init__()
        self.extractor = RustElementExtractor()
        self.language = "rust"
        self.supported_extensions = self.get_file_extensions()
        self._cached_language: Any | None = None

    def get_language_name(self) -> str:
        """Get the language name."""
        return "rust"

    def get_file_extensions(self) -> list[str]:
        """Get supported file extensions."""
        return [".rs"]

    def create_extractor(self) -> ElementExtractor:
        """Create a new element extractor instance."""
        return RustElementExtractor()

    async def analyze_file(
        self, file_path: str, request: "AnalysisRequest"
    ) -> "AnalysisResult":
        """Analyze Rust code and return structured results."""

        from ..models import AnalysisResult

        try:
            from ..encoding_utils import read_file_safe

            file_content, detected_encoding = read_file_safe(file_path)

            # Get tree-sitter language and parse
            language = self.get_tree_sitter_language()
            if language is None:
                return AnalysisResult(
                    file_path=file_path,
                    language="rust",
                    line_count=len(file_content.split("\n")),
                    elements=[],
                    source_code=file_content,
                )

            import tree_sitter

            parser = tree_sitter.Parser()

            # Set language
            if hasattr(parser, "set_language"):
                parser.set_language(language)
            elif hasattr(parser, "language"):
                parser.language = language
            else:
                parser = tree_sitter.Parser(language)

            tree = parser.parse(file_content.encode("utf-8"))

            # Extract elements
            elements_dict = self.extract_elements(tree, file_content)

            all_elements = []
            all_elements.extend(elements_dict.get("functions", []))
            all_elements.extend(elements_dict.get("classes", []))
            all_elements.extend(elements_dict.get("variables", []))
            all_elements.extend(elements_dict.get("imports", []))

            # Count nodes
            node_count = (
                self._count_tree_nodes(tree.root_node) if tree and tree.root_node else 0
            )

            # Pass extra Rust metadata (impls, modules) via result object or merged into elements
            # For now, we rely on the standard AnalysisResult fields, but the formatter needs impls/modules.
            # We can attach them to the AnalysisResult object dynamically or put them in elements list if they are CodeElements.
            # Currently AnalysisResult.elements is list[CodeElement].
            # We can't easily add dicts to elements list if they are not CodeElements.
            # But the formatter receives `analysis_result: dict[str, Any]` which is `as_dict()` of AnalysisResult.
            # We need to ensure `as_dict()` includes our extra data if we add it as attributes.

            result = AnalysisResult(
                file_path=file_path,
                language="rust",
                line_count=len(file_content.split("\n")),
                elements=all_elements,
                node_count=node_count,
                source_code=file_content,
            )

            # Attach extra metadata for the formatter
            # Note: This requires AnalysisResult to handle arbitrary attributes or `as_dict` to include them.
            # Looking at models.py (not visible here but assumed), standard attributes are fixed.
            # However, `AnalysisResult` might allow dynamic attributes.
            # Alternatively, we include them as custom elements in `elements` list.

            # Let's check models.py later. For now, we'll attach them and hope for the best or modify models.py if needed.
            result.modules = self.extractor.modules
            result.impls = self.extractor.impl_blocks

            return result

        except Exception as e:
            log_error(f"Error analyzing Rust file {file_path}: {e}")
            return AnalysisResult(
                file_path=file_path,
                language="rust",
                line_count=0,
                elements=[],
                source_code="",
                error_message=str(e),
                success=False,
            )

    def _count_tree_nodes(self, node: Any) -> int:
        """Recursively count nodes."""
        if node is None:
            return 0
        count = 1
        if hasattr(node, "children"):
            for child in node.children:
                count += self._count_tree_nodes(child)
        return count

    def get_tree_sitter_language(self) -> Any | None:
        """Get the tree-sitter language for Rust."""
        if self._cached_language is not None:
            return self._cached_language

        try:
            import tree_sitter
            import tree_sitter_rust

            caps_or_lang = tree_sitter_rust.language()

            if hasattr(caps_or_lang, "__class__") and "Language" in str(
                type(caps_or_lang)
            ):
                self._cached_language = caps_or_lang
            else:
                try:
                    self._cached_language = tree_sitter.Language(caps_or_lang)
                except Exception as e:
                    log_error(f"Failed to create Language object: {e}")
                    return None

            return self._cached_language
        except ImportError as e:
            log_error(f"tree-sitter-rust not available: {e}")
            return None
        except Exception as e:
            log_error(f"Failed to load tree-sitter language for Rust: {e}")
            return None

    def extract_elements(self, tree: Any | None, source_code: str) -> dict[str, Any]:
        """Extract all elements."""
        if tree is None:
            return {"functions": [], "classes": [], "variables": []}

        try:
            # Reset extractor state
            # We need to ensure we use the same extractor instance to collect side-effects like modules/impls
            # But create_extractor() creates a new one.
            # Here we use self.extractor which is initialized in __init__
            # Wait, `analyze_file` calls `extract_elements` which calls `create_extractor` in Java plugin...
            # In JavaPlugin.extract_elements: `extractor = self.create_extractor()`
            # This means new extractor every time.
            # So if we want to access `modules` and `impls` after extraction, we need to get them from THAT extractor instance.

            extractor = (
                self.create_extractor()
            )  # Create new instance for thread safety / isolation

            result = {
                "functions": extractor.extract_functions(tree, source_code),
                "classes": extractor.extract_classes(tree, source_code),
                "variables": extractor.extract_variables(tree, source_code),
                "imports": extractor.extract_imports(tree, source_code),
            }

            # Capture side-effects
            if isinstance(extractor, RustElementExtractor):
                self.extractor.modules = extractor.modules
                self.extractor.impl_blocks = extractor.impl_blocks

            return result

        except Exception as e:
            log_error(f"Error extracting elements: {e}")
            return {"functions": [], "classes": [], "variables": []}

    def supports_file(self, file_path: str) -> bool:
        """Check if this plugin supports the given file."""
        return any(
            file_path.lower().endswith(ext) for ext in self.get_file_extensions()
        )



#!/usr/bin/env python3
"""
Project Root Detection Demo

Demonstrates the intelligent project root detection and configuration
across CLI and MCP interfaces.
"""

import asyncio
import tempfile
from pathlib import Path

from tree_sitter_analyzer.core.analysis_engine import get_analysis_engine
from tree_sitter_analyzer.mcp.tools.table_format_tool import TableFormatTool
from tree_sitter_analyzer.project_detector import (
    ProjectRootDetector,
    detect_project_root,
)


async def main():
    """Demonstrate project root detection and usage."""
    print("🔍 Tree-sitter Analyzer Project Root Detection Demo")
    print("=" * 70)

    # Create a mock project structure
    temp_dir = Path(tempfile.mkdtemp())
    project_root = temp_dir / "demo_project"

    # Create project structure
    project_root.mkdir()
    (project_root / "src").mkdir()
    (project_root / "tests").mkdir()

    # Create project markers
    with open(project_root / "pyproject.toml", "w") as f:
        f.write(
            """
[tool.poetry]
name = "demo-project"
version = "0.1.0"
description = "Demo project for tree-sitter-analyzer"

[tool.poetry.dependencies]
python = "^3.8"
"""
        )

    with open(project_root / "README.md", "w") as f:
        f.write(
            "# Demo Project\n\nThis is a demo project for testing project root detection."
        )

    # Create source files
    src_file = project_root / "src" / "main.py"
    with open(src_file, "w") as f:
        f.write(
            """
def fibonacci(n):
    '''Calculate fibonacci number'''
    if n <= 1:
        return n
    return fibonacci(n-1) + fibonacci(n-2)

class Calculator:
    def __init__(self):
        self.history = []

    def add(self, a, b):
        result = a + b
        self.history.append(f"{a} + {b} = {result}")
        return result

    def multiply(self, a, b):
        result = a * b
        self.history.append(f"{a} * {b} = {result}")
        return result
"""
        )

    test_file = project_root / "tests" / "test_main.py"
    with open(test_file, "w") as f:
        f.write(
            """
import unittest
from src.main import Calculator, fibonacci

class TestCalculator(unittest.TestCase):
    def setUp(self):
        self.calc = Calculator()

    def test_add(self):
        self.assertEqual(self.calc.add(2, 3), 5)

    def test_multiply(self):
        self.assertEqual(self.calc.multiply(4, 5), 20)

class TestFibonacci(unittest.TestCase):
    def test_fibonacci(self):
        self.assertEqual(fibonacci(0), 0)
        self.assertEqual(fibonacci(1), 1)
        self.assertEqual(fibonacci(5), 5)
"""
        )

    print(f"📁 Created demo project at: {project_root}")
    print(f"📄 Source file: {src_file}")
    print(f"🧪 Test file: {test_file}")
    print()

    # 1. Test Project Root Detection
    print("1️⃣ Testing Project Root Detection")
    print("-" * 50)

    detector = ProjectRootDetector()

    # Test detection from source file
    detected_from_src = detector.detect_from_file(str(src_file))
    print(f"✅ Detection from src file: {detected_from_src}")
    print(f"   Expected: {project_root}")
    print(
        f"   Match: {'✅ YES' if detected_from_src == str(project_root) else '❌ NO'}"
    )

    # Test detection from test file
    detected_from_test = detector.detect_from_file(str(test_file))
    print(f"✅ Detection from test file: {detected_from_test}")
    print(
        f"   Match: {'✅ YES' if detected_from_test == str(project_root) else '❌ NO'}"
    )

    print()

    # 2. Test Unified Detection Function
    print("2️⃣ Testing Unified Detection Function")
    print("-" * 50)

    # Test with file path only
    unified_result1 = detect_project_root(str(src_file))
    print(f"📍 Auto-detection from file: {unified_result1}")

    # Test with explicit root
    explicit_root = temp_dir / "custom_root"
    explicit_root.mkdir()
    unified_result2 = detect_project_root(str(src_file), str(explicit_root))
    print(f"📍 With explicit root: {unified_result2}")
    print(
        f"   Uses explicit: {'✅ YES' if unified_result2 == str(explicit_root.resolve()) else '❌ NO'}"
    )

    print()

    # 3. Test Analysis Engine Integration
    print("3️⃣ Testing Analysis Engine Integration")
    print("-" * 50)

    # Test with auto-detected project root
    engine1 = get_analysis_engine(detected_from_src)
    engine1_root = getattr(engine1, "_project_root", "None")
    print(f"🔧 Engine with auto-detected root: {engine1_root}")

    # Test with explicit project root
    engine2 = get_analysis_engine(str(explicit_root))
    engine2_root = getattr(engine2, "_project_root", "None")
    print(f"🔧 Engine with explicit root: {engine2_root}")

    # Verify they are different instances (due to different project roots)
    print(f"🔧 Different instances: {'✅ YES' if engine1 is not engine2 else '❌ NO'}")

    print()

    # 4. Test MCP Tool Integration
    print("4️⃣ Testing MCP Tool Integration")
    print("-" * 50)

    # Test MCP tool with auto-detected project root
    mcp_tool1 = TableFormatTool(detected_from_src)
    tool1_root = (
        mcp_tool1.security_validator.boundary_manager.project_root
        if mcp_tool1.security_validator.boundary_manager
        else "None"
    )
    print(f"🛠️ MCP tool with auto-detected root: {tool1_root}")

    # Test MCP tool with explicit project root
    mcp_tool2 = TableFormatTool(str(explicit_root))
    tool2_root = (
        mcp_tool2.security_validator.boundary_manager.project_root
        if mcp_tool2.security_validator.boundary_manager
        else "None"
    )
    print(f"🛠️ MCP tool with explicit root: {tool2_root}")

    # Test actual analysis
    try:
        result = await mcp_tool1.execute({"file_path": str(src_file)})
        print(
            f"✅ Analysis successful: Found {len(result.get('elements', []))} elements"
        )
    except Exception as e:
        print(f"❌ Analysis failed: {e}")

    print()

    # 5. Test Security Validation
    print("5️⃣ Testing Security Validation")
    print("-" * 50)

    # Test file within project
    is_valid1, msg1 = mcp_tool1.security_validator.validate_file_path(str(src_file))
    print(
        f"🔒 File within project: {'✅ VALID' if is_valid1 else '❌ INVALID'} - {msg1}"
    )

    # Test file outside project
    outside_file = temp_dir / "outside.py"
    with open(outside_file, "w") as f:
        f.write("print('outside')")

    is_valid2, msg2 = mcp_tool1.security_validator.validate_file_path(str(outside_file))
    print(
        f"🔒 File outside project: {'✅ VALID' if is_valid2 else '❌ INVALID'} - {msg2}"
    )

    # Test malicious path
    is_valid3, msg3 = mcp_tool1.security_validator.validate_file_path(
        "../../../etc/passwd"
    )
    print(
        f"🔒 Malicious path: {'✅ BLOCKED' if not is_valid3 else '❌ ALLOWED'} - {msg3}"
    )

    print()

    # 6. Test CLI Integration Simulation
    print("6️⃣ CLI Integration Examples")
    print("-" * 50)

    print("📝 CLI Usage Examples:")
    print("   # Auto-detect project root:")
    print(f"   tree-sitter-analyzer {src_file} --table=full")
    print()
    print("   # Explicit project root:")
    print(
        f"   tree-sitter-analyzer {src_file} --project-root {project_root} --table=full"
    )
    print()

    # 7. Test MCP Server Integration Simulation
    print("7️⃣ MCP Server Integration Examples")
    print("-" * 50)

    print("📝 MCP Server Configuration Examples:")
    print("   # Auto-detect project root:")
    print('   "command": "python", "args": ["-m", "tree_sitter_analyzer.mcp.server"]')
    print()
    print("   # Explicit project root:")
    print(
        f'   "command": "python", "args": ["-m", "tree_sitter_analyzer.mcp.server", "--project-root", "{project_root}"]'
    )
    print()
    print("   # Environment variable:")
    print(f'   "env": {{"TREE_SITTER_PROJECT_ROOT": "{project_root}"}}')
    print()

    # Cleanup
    import shutil

    shutil.rmtree(str(temp_dir), ignore_errors=True)

    print("🎉 Project Root Detection Demo Complete!")
    print("=" * 70)
    print("Summary:")
    print("✅ Intelligent project root detection from common markers")
    print("✅ Priority-based configuration (explicit > auto-detect > fallback)")
    print("✅ CLI support with --project-root parameter")
    print("✅ MCP server support with command line arguments")
    print("✅ Environment variable support (TREE_SITTER_PROJECT_ROOT)")
    print("✅ Security validation with project boundaries")
    print("✅ Multiple project root support (different instances)")
    print()
    print("🚀 Your tree-sitter-analyzer now has intelligent project management!")


if __name__ == "__main__":
    asyncio.run(main())



#!/usr/bin/env python
"""Run full coverage report and save results."""

import json
import os
import subprocess
import sys


def run_quick_test():
    """Run quick test without coverage to verify tests pass."""
    print("=" * 60)
    print("Running quick test (no coverage) to verify all tests pass...")
    print("=" * 60)

    cmd = [sys.executable, "-m", "pytest", "tests/", "-q", "--tb=no"]

    result = subprocess.run(
        cmd,
        capture_output=True,
        text=True,
        cwd=os.path.dirname(os.path.abspath(__file__)),
    )

    # Write output
    with open("test_output.txt", "w", encoding="utf-8") as f:
        f.write(result.stdout)
        f.write("\n")
        f.write(result.stderr)

    # Print summary
    lines = result.stdout.strip().split("\n")
    print("\n".join(lines[-20:]))

    return result.returncode


def run_coverage():
    """Run pytest with coverage."""
    print("=" * 60)
    print("Running full test suite with coverage...")
    print("=" * 60)

    cmd = [
        sys.executable,
        "-m",
        "pytest",
        "tests/",
        "--cov=tree_sitter_analyzer",
        "--cov-branch",
        "--cov-report=term",
        "--cov-report=json",
        "--cov-report=html",
        "-q",
        "--tb=no",
    ]

    print(f"Command: {' '.join(cmd)}")

    result = subprocess.run(
        cmd,
        capture_output=True,
        text=True,
        cwd=os.path.dirname(os.path.abspath(__file__)),
    )

    # Write output to file
    with open("coverage_output.txt", "w", encoding="utf-8") as f:
        f.write("=== STDOUT ===\n")
        f.write(result.stdout)
        f.write("\n\n=== STDERR ===\n")
        f.write(result.stderr)
        f.write(f"\n\n=== RETURN CODE: {result.returncode} ===\n")

    print(f"Return code: {result.returncode}")
    print("Output written to coverage_output.txt")

    # Print last 200 lines of stdout
    lines = result.stdout.split("\n")
    print("\n--- Last 200 lines of output ---")
    for line in lines[-200:]:
        print(line)

    return result.returncode


def analyze_coverage():
    """Analyze coverage.json if it exists."""
    print("\n" + "=" * 60)
    print("Analyzing coverage results...")
    print("=" * 60)

    if not os.path.exists("coverage.json"):
        print("coverage.json not found!")
        return

    with open("coverage.json") as f:
        data = json.load(f)

    totals = data.get("totals", {})
    print("\nOverall Coverage:")
    print(f"  Line Coverage: {totals.get('percent_covered', 0):.2f}%")
    print(f"  Covered Lines: {totals.get('covered_lines', 0)}")
    print(f"  Total Lines: {totals.get('num_statements', 0)}")
    print(f"  Missing Lines: {totals.get('missing_lines', 0)}")
    print(f"  Excluded Lines: {totals.get('excluded_lines', 0)}")

    if "covered_branches" in totals:
        print("\nBranch Coverage:")
        print(f"  Covered Branches: {totals.get('covered_branches', 0)}")
        print(f"  Total Branches: {totals.get('num_branches', 0)}")
        print(f"  Missing Branches: {totals.get('missing_branches', 0)}")

    # Analyze by file
    files = data.get("files", {})
    low_coverage = []
    high_coverage = []

    for path, file_data in files.items():
        summary = file_data.get("summary", {})
        percent = summary.get("percent_covered", 0)
        statements = summary.get("num_statements", 0)

        if statements > 20:
            if percent < 50:
                low_coverage.append((path, percent, statements))
            elif percent >= 80:
                high_coverage.append((path, percent, statements))

    print(f"\n\nHigh Coverage Files (>= 80%, > 20 statements): {len(high_coverage)}")
    for path, pct, stmts in sorted(high_coverage, key=lambda x: -x[1])[:10]:
        print(f"  {pct:.1f}% - {os.path.basename(path)} ({stmts} stmts)")

    print(f"\n\nLow Coverage Files (< 50%, > 20 statements): {len(low_coverage)}")
    for path, pct, stmts in sorted(low_coverage, key=lambda x: x[1])[:15]:
        print(f"  {pct:.1f}% - {os.path.basename(path)} ({stmts} stmts)")


def main():
    import sys

    if len(sys.argv) > 1 and sys.argv[1] == "--quick":
        return run_quick_test()
    elif len(sys.argv) > 1 and sys.argv[1] == "--analyze":
        analyze_coverage()
        return 0
    else:
        rc = run_coverage()
        if rc == 0:
            analyze_coverage()
        return rc


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
"""
Format Change Management System

Manages format changes, approvals, tracking, and rollback capabilities.
Ensures controlled evolution of format specifications.
"""

import argparse
import hashlib
import json
import logging
import shutil
import sqlite3
import sys
from datetime import datetime
from pathlib import Path
from typing import Any

# Add project root to path - needs to be before imports  # noqa: E402
project_root = Path(__file__).parent.parent
sys.path.insert(0, str(project_root))

from tests.format_testing.golden_master import GoldenMasterManager  # noqa: E402


class FormatChangeDatabase:
    """Database for tracking format changes"""

    def __init__(self, db_path: str):
        self.db_path = db_path
        self.init_database()

    def init_database(self):
        """Initialize change management database"""
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()

        # Format change requests table
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS format_change_requests (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                timestamp TEXT NOT NULL,
                requester TEXT NOT NULL,
                format_type TEXT NOT NULL,
                change_type TEXT NOT NULL,
                description TEXT NOT NULL,
                justification TEXT NOT NULL,
                impact_assessment TEXT,
                status TEXT DEFAULT 'pending',
                reviewer TEXT,
                review_timestamp TEXT,
                review_comments TEXT,
                implementation_timestamp TEXT,
                rollback_timestamp TEXT
            )
        """
        )

        # Format versions table
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS format_versions (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                format_type TEXT NOT NULL,
                version TEXT NOT NULL,
                timestamp TEXT NOT NULL,
                change_request_id INTEGER,
                specification_hash TEXT NOT NULL,
                golden_master_hash TEXT,
                is_active BOOLEAN DEFAULT TRUE,
                rollback_version TEXT,
                FOREIGN KEY (change_request_id) REFERENCES format_change_requests (id)
            )
        """
        )

        # Format compatibility matrix table
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS format_compatibility (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                format_type TEXT NOT NULL,
                from_version TEXT NOT NULL,
                to_version TEXT NOT NULL,
                compatibility_level TEXT NOT NULL,
                breaking_changes TEXT,
                migration_required BOOLEAN DEFAULT FALSE,
                migration_script TEXT
            )
        """
        )

        # Format approval workflow table
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS format_approvals (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                change_request_id INTEGER NOT NULL,
                approver TEXT NOT NULL,
                approval_timestamp TEXT NOT NULL,
                approval_status TEXT NOT NULL,
                comments TEXT,
                FOREIGN KEY (change_request_id) REFERENCES format_change_requests (id)
            )
        """
        )

        conn.commit()
        conn.close()

    def create_change_request(
        self,
        requester: str,
        format_type: str,
        change_type: str,
        description: str,
        justification: str,
        impact_assessment: str = "",
    ) -> int:
        """Create a new format change request"""
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()

        cursor.execute(
            """
            INSERT INTO format_change_requests
            (timestamp, requester, format_type, change_type, description,
             justification, impact_assessment)
            VALUES (?, ?, ?, ?, ?, ?, ?)
        """,
            (
                datetime.utcnow().isoformat(),
                requester,
                format_type,
                change_type,
                description,
                justification,
                impact_assessment,
            ),
        )

        change_request_id = cursor.lastrowid
        conn.commit()
        conn.close()

        return change_request_id

    def add_approval(
        self, change_request_id: int, approver: str, status: str, comments: str = ""
    ):
        """Add approval to change request"""
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()

        cursor.execute(
            """
            INSERT INTO format_approvals
            (change_request_id, approver, approval_timestamp, approval_status, comments)
            VALUES (?, ?, ?, ?, ?)
        """,
            (
                change_request_id,
                approver,
                datetime.utcnow().isoformat(),
                status,
                comments,
            ),
        )

        conn.commit()
        conn.close()

    def update_change_request_status(
        self,
        change_request_id: int,
        status: str,
        reviewer: str = "",
        comments: str = "",
    ):
        """Update change request status"""
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()

        cursor.execute(
            """
            UPDATE format_change_requests
            SET status = ?, reviewer = ?, review_timestamp = ?, review_comments = ?
            WHERE id = ?
        """,
            (
                status,
                reviewer,
                datetime.utcnow().isoformat(),
                comments,
                change_request_id,
            ),
        )

        conn.commit()
        conn.close()

    def create_format_version(
        self,
        format_type: str,
        version: str,
        change_request_id: int,
        specification_hash: str,
        golden_master_hash: str = "",
    ) -> int:
        """Create new format version"""
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()

        # Deactivate previous versions
        cursor.execute(
            """
            UPDATE format_versions
            SET is_active = FALSE
            WHERE format_type = ? AND is_active = TRUE
        """,
            (format_type,),
        )

        # Create new version
        cursor.execute(
            """
            INSERT INTO format_versions
            (format_type, version, timestamp, change_request_id,
             specification_hash, golden_master_hash, is_active)
            VALUES (?, ?, ?, ?, ?, ?, TRUE)
        """,
            (
                format_type,
                version,
                datetime.utcnow().isoformat(),
                change_request_id,
                specification_hash,
                golden_master_hash,
            ),
        )

        version_id = cursor.lastrowid
        conn.commit()
        conn.close()

        return version_id

    def get_pending_requests(self) -> list[dict[str, Any]]:
        """Get pending change requests"""
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()

        cursor.execute(
            """
            SELECT * FROM format_change_requests
            WHERE status = 'pending'
            ORDER BY timestamp DESC
        """
        )

        columns = [desc[0] for desc in cursor.description]
        results = [dict(zip(columns, row, strict=False)) for row in cursor.fetchall()]

        conn.close()
        return results

    def get_change_request(self, request_id: int) -> dict[str, Any] | None:
        """Get specific change request"""
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()

        cursor.execute(
            """
            SELECT * FROM format_change_requests WHERE id = ?
        """,
            (request_id,),
        )

        row = cursor.fetchone()
        if row:
            columns = [desc[0] for desc in cursor.description]
            result = dict(zip(columns, row, strict=False))
        else:
            result = None

        conn.close()
        return result

    def get_format_versions(self, format_type: str) -> list[dict[str, Any]]:
        """Get format version history"""
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()

        cursor.execute(
            """
            SELECT * FROM format_versions
            WHERE format_type = ?
            ORDER BY timestamp DESC
        """,
            (format_type,),
        )

        columns = [desc[0] for desc in cursor.description]
        results = [dict(zip(columns, row, strict=False)) for row in cursor.fetchall()]

        conn.close()
        return results

    def get_active_version(self, format_type: str) -> dict[str, Any] | None:
        """Get active format version"""
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()

        cursor.execute(
            """
            SELECT * FROM format_versions
            WHERE format_type = ? AND is_active = TRUE
        """,
            (format_type,),
        )

        row = cursor.fetchone()
        if row:
            columns = [desc[0] for desc in cursor.description]
            result = dict(zip(columns, row, strict=False))
        else:
            result = None

        conn.close()
        return result


class FormatChangeManager:
    """Format change management system"""

    def __init__(self, db_path: str = "format_changes.db"):
        self.db = FormatChangeDatabase(db_path)
        self.golden_master_manager = GoldenMasterManager()
        self.logger = self._setup_logging()

        # Change type definitions
        self.change_types = {
            "enhancement": "Format enhancement or new feature",
            "bugfix": "Fix for format output issues",
            "breaking": "Breaking change requiring migration",
            "deprecation": "Deprecation of format features",
            "security": "Security-related format changes",
        }

        # Approval requirements by change type
        self.approval_requirements = {
            "enhancement": ["tech_lead"],
            "bugfix": ["developer"],
            "breaking": ["tech_lead", "product_owner"],
            "deprecation": ["tech_lead", "product_owner"],
            "security": ["security_lead", "tech_lead"],
        }

    def _setup_logging(self) -> logging.Logger:
        """Setup logging for change management"""
        logger = logging.getLogger("format_change_manager")
        logger.setLevel(logging.INFO)

        handler = logging.StreamHandler()
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        handler.setFormatter(formatter)
        logger.addHandler(handler)

        return logger

    def create_change_request(
        self,
        requester: str,
        format_type: str,
        change_type: str,
        description: str,
        justification: str,
        impact_assessment: str = "",
    ) -> int:
        """Create new format change request"""
        if change_type not in self.change_types:
            raise ValueError(f"Invalid change type: {change_type}")

        if format_type not in ["full", "compact", "csv"]:
            raise ValueError(f"Invalid format type: {format_type}")

        # Perform impact assessment if not provided
        if not impact_assessment:
            impact_assessment = self._assess_change_impact(
                format_type, change_type, description
            )

        request_id = self.db.create_change_request(
            requester,
            format_type,
            change_type,
            description,
            justification,
            impact_assessment,
        )

        self.logger.info(
            f"Created change request {request_id} for {format_type} format"
        )
        return request_id

    def _assess_change_impact(
        self, format_type: str, change_type: str, description: str
    ) -> str:
        """Assess impact of proposed change"""
        impact_factors = []

        # Assess based on change type
        if change_type == "breaking":
            impact_factors.append("HIGH: Breaking change requires migration")
        elif change_type == "security":
            impact_factors.append("HIGH: Security implications")
        elif change_type == "deprecation":
            impact_factors.append("MEDIUM: Deprecation affects existing users")
        else:
            impact_factors.append("LOW: Non-breaking change")

        # Assess based on format type
        if format_type == "csv":
            impact_factors.append("MEDIUM: CSV format changes affect data processing")
        elif format_type == "full":
            impact_factors.append("HIGH: Full format is primary output format")
        else:
            impact_factors.append("LOW: Compact format has limited usage")

        # Check for keywords in description
        high_impact_keywords = ["schema", "structure", "header", "column", "field"]
        if any(keyword in description.lower() for keyword in high_impact_keywords):
            impact_factors.append("HIGH: Structural changes detected")

        return "; ".join(impact_factors)

    def approve_change_request(
        self, request_id: int, approver: str, status: str, comments: str = ""
    ):
        """Approve or reject change request"""
        if status not in ["approved", "rejected", "needs_revision"]:
            raise ValueError(f"Invalid approval status: {status}")

        request = self.db.get_change_request(request_id)
        if not request:
            raise ValueError(f"Change request {request_id} not found")

        # Add approval record
        self.db.add_approval(request_id, approver, status, comments)

        # Check if all required approvals are met
        if status == "approved":
            required_approvers = self.approval_requirements.get(
                request["change_type"], ["developer"]
            )
            if self._check_approval_requirements(request_id, required_approvers):
                self.db.update_change_request_status(
                    request_id, "approved", approver, comments
                )
                self.logger.info(f"Change request {request_id} fully approved")
            else:
                self.logger.info(
                    f"Change request {request_id} partially approved, waiting for more approvals"
                )
        else:
            self.db.update_change_request_status(request_id, status, approver, comments)
            self.logger.info(f"Change request {request_id} {status}")

    def _check_approval_requirements(
        self, request_id: int, required_approvers: list[str]
    ) -> bool:
        """Check if approval requirements are met"""
        # This is a simplified check - in practice, you'd check actual approver roles
        # For now, we'll assume any approval from required roles is sufficient
        return True

    def implement_change(self, request_id: int, implementer: str) -> str:
        """Implement approved format change"""
        request = self.db.get_change_request(request_id)
        if not request:
            raise ValueError(f"Change request {request_id} not found")

        if request["status"] != "approved":
            raise ValueError(f"Change request {request_id} is not approved")

        format_type = request["format_type"]

        # Generate new version number
        versions = self.db.get_format_versions(format_type)
        if versions:
            latest_version = versions[0]["version"]
            major, minor, patch = map(int, latest_version.split("."))

            if request["change_type"] == "breaking":
                new_version = f"{major + 1}.0.0"
            elif request["change_type"] in ["enhancement", "deprecation"]:
                new_version = f"{major}.{minor + 1}.0"
            else:
                new_version = f"{major}.{minor}.{patch + 1}"
        else:
            new_version = "1.0.0"

        # Create backup of current format specification
        self._backup_current_format(format_type, new_version)

        # Calculate specification hash (simplified - would hash actual spec file)
        spec_hash = hashlib.md5(
            f"{format_type}_{new_version}_{datetime.utcnow()}".encode()
        ).hexdigest()

        # Create new format version
        self.db.create_format_version(format_type, new_version, request_id, spec_hash)

        # Update change request status
        self.db.update_change_request_status(request_id, "implemented", implementer)

        self.logger.info(
            f"Implemented change request {request_id} as version {new_version}"
        )

        return new_version

    def _backup_current_format(self, format_type: str, new_version: str) -> str:
        """Backup current format specification"""
        backup_dir = Path("format_backups")
        backup_dir.mkdir(exist_ok=True)

        timestamp = datetime.utcnow().strftime("%Y%m%d_%H%M%S")
        backup_path = backup_dir / f"{format_type}_backup_{timestamp}_v{new_version}"
        backup_path.mkdir(exist_ok=True)

        # Backup golden masters
        golden_master_dir = Path("tests/format_testing/golden_masters")
        if golden_master_dir.exists():
            shutil.copytree(
                golden_master_dir, backup_path / "golden_masters", dirs_exist_ok=True
            )

        # Backup format specifications
        spec_file = Path("docs/format_specifications.md")
        if spec_file.exists():
            shutil.copy2(spec_file, backup_path / "format_specifications.md")

        self.logger.info(f"Created backup at {backup_path}")
        return str(backup_path)

    def rollback_change(self, request_id: int, rollback_reason: str) -> str:
        """Rollback implemented format change"""
        request = self.db.get_change_request(request_id)
        if not request:
            raise ValueError(f"Change request {request_id} not found")

        if request["status"] != "implemented":
            raise ValueError(f"Change request {request_id} is not implemented")

        format_type = request["format_type"]

        # Find the version to rollback to
        versions = self.db.get_format_versions(format_type)
        if len(versions) < 2:
            raise ValueError("No previous version to rollback to")

        current_version = versions[0]
        previous_version = versions[1]

        # Restore from backup
        self._restore_from_backup(format_type, previous_version["version"])

        # Create rollback version entry
        rollback_version = f"{current_version['version']}-rollback"
        self.db.create_format_version(
            format_type,
            rollback_version,
            request_id,
            previous_version["specification_hash"],
            previous_version["golden_master_hash"],
        )

        # Update change request
        self.db.update_change_request_status(
            request_id, "rolled_back", "", rollback_reason
        )

        self.logger.info(
            f"Rolled back change request {request_id} to version {previous_version['version']}"
        )

        return rollback_version

    def _restore_from_backup(self, format_type: str, version: str):
        """Restore format from backup"""
        backup_dir = Path("format_backups")

        # Find the most recent backup for this version
        backup_pattern = f"{format_type}_backup_*_v{version}"
        backups = list(backup_dir.glob(backup_pattern))

        if not backups:
            raise ValueError(f"No backup found for {format_type} version {version}")

        latest_backup = max(backups, key=lambda p: p.stat().st_mtime)

        # Restore golden masters
        golden_master_backup = latest_backup / "golden_masters"
        if golden_master_backup.exists():
            golden_master_dir = Path("tests/format_testing/golden_masters")
            if golden_master_dir.exists():
                shutil.rmtree(golden_master_dir)
            shutil.copytree(golden_master_backup, golden_master_dir)

        # Restore format specifications
        spec_backup = latest_backup / "format_specifications.md"
        if spec_backup.exists():
            shutil.copy2(spec_backup, "docs/format_specifications.md")

        self.logger.info(f"Restored from backup {latest_backup}")

    def generate_change_report(self, output_file: str = "format_change_report.json"):
        """Generate comprehensive change management report"""
        pending_requests = self.db.get_pending_requests()

        # Get version history for all formats
        format_histories = {}
        for format_type in ["full", "compact", "csv"]:
            format_histories[format_type] = self.db.get_format_versions(format_type)

        report = {
            "timestamp": datetime.utcnow().isoformat(),
            "summary": {
                "pending_requests": len(pending_requests),
                "total_formats": len(format_histories),
                "active_versions": {
                    fmt: versions[0]["version"] if versions else "none"
                    for fmt, versions in format_histories.items()
                },
            },
            "pending_requests": pending_requests,
            "format_histories": format_histories,
            "change_types": self.change_types,
            "approval_requirements": self.approval_requirements,
        }

        with open(output_file, "w") as f:
            json.dump(report, f, indent=2, default=str)

        self.logger.info(f"Generated change report: {output_file}")
        return report


def main():
    """Main change management function"""
    parser = argparse.ArgumentParser(description="Format change management")
    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    # Create change request
    create_parser = subparsers.add_parser("create", help="Create change request")
    create_parser.add_argument("--requester", required=True, help="Requester name")
    create_parser.add_argument(
        "--format",
        required=True,
        choices=["full", "compact", "csv"],
        help="Format type",
    )
    create_parser.add_argument(
        "--type",
        required=True,
        choices=["enhancement", "bugfix", "breaking", "deprecation", "security"],
        help="Change type",
    )
    create_parser.add_argument(
        "--description", required=True, help="Change description"
    )
    create_parser.add_argument(
        "--justification", required=True, help="Change justification"
    )
    create_parser.add_argument("--impact", help="Impact assessment")

    # Approve change request
    approve_parser = subparsers.add_parser("approve", help="Approve change request")
    approve_parser.add_argument(
        "--request-id", type=int, required=True, help="Request ID"
    )
    approve_parser.add_argument("--approver", required=True, help="Approver name")
    approve_parser.add_argument(
        "--status",
        required=True,
        choices=["approved", "rejected", "needs_revision"],
        help="Approval status",
    )
    approve_parser.add_argument("--comments", help="Approval comments")

    # Implement change
    implement_parser = subparsers.add_parser("implement", help="Implement change")
    implement_parser.add_argument(
        "--request-id", type=int, required=True, help="Request ID"
    )
    implement_parser.add_argument(
        "--implementer", required=True, help="Implementer name"
    )

    # Rollback change
    rollback_parser = subparsers.add_parser("rollback", help="Rollback change")
    rollback_parser.add_argument(
        "--request-id", type=int, required=True, help="Request ID"
    )
    rollback_parser.add_argument("--reason", required=True, help="Rollback reason")

    # List pending requests
    subparsers.add_parser("list", help="List pending requests")

    # Generate report
    report_parser = subparsers.add_parser("report", help="Generate change report")
    report_parser.add_argument(
        "--output", default="format_change_report.json", help="Output file"
    )

    # Database path
    parser.add_argument("--db-path", default="format_changes.db", help="Database path")

    args = parser.parse_args()

    if not args.command:
        parser.print_help()
        return 1

    manager = FormatChangeManager(args.db_path)

    try:
        if args.command == "create":
            request_id = manager.create_change_request(
                args.requester,
                args.format,
                args.type,
                args.description,
                args.justification,
                args.impact or "",
            )
            print(f"Created change request {request_id}")

        elif args.command == "approve":
            manager.approve_change_request(
                args.request_id, args.approver, args.status, args.comments or ""
            )
            print(f"Updated approval for request {args.request_id}")

        elif args.command == "implement":
            version = manager.implement_change(args.request_id, args.implementer)
            print(f"Implemented change request {args.request_id} as version {version}")

        elif args.command == "rollback":
            version = manager.rollback_change(args.request_id, args.reason)
            print(f"Rolled back change request {args.request_id} to {version}")

        elif args.command == "list":
            pending = manager.db.get_pending_requests()
            if pending:
                print(f"Pending change requests ({len(pending)}):")
                for req in pending:
                    print(
                        f"  {req['id']}: {req['format_type']} - {req['change_type']} - {req['description'][:50]}..."
                    )
            else:
                print("No pending change requests")

        elif args.command == "report":
            report = manager.generate_change_report(args.output)
            print(f"Generated change report: {args.output}")
            print(f"Pending requests: {report['summary']['pending_requests']}")

        return 0

    except Exception as e:
        print(f"Error: {e}")
        return 1


if __name__ == "__main__":
    sys.exit(main())
